# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

"""
Build a combined news file from news fragments.
"""

from __future__ import annotations

import os
import sys

from datetime import date
from pathlib import Path

import click

from click import Context, Option, UsageError

from towncrier import _git

from ._builder import find_fragments, render_fragments, split_fragments
from ._project import get_project_name, get_version
from ._settings import ConfigError, config_option_help, load_config_from_options
from ._writer import append_to_newsfile


if sys.version_info < (3, 10):
    import importlib_resources as resources
else:
    from importlib import resources


def _get_date() -> str:
    return date.today().isoformat()


def _validate_answer(ctx: Context, param: Option, value: bool) -> bool:
    value_check = (
        ctx.params.get("answer_yes")
        if param.name == "answer_keep"
        else ctx.params.get("answer_keep")
    )
    if value_check and value:
        click.echo("You can not choose both --yes and --keep at the same time")
        ctx.abort()
    return value


@click.command(name="build")
@click.option(
    "--draft",
    "draft",
    default=False,
    flag_value=True,
    help=(
        "Render the news fragments to standard output. "
        "Don't write to files, don't check versions."
    ),
)
@click.option(
    "--config",
    "config_file",
    default=None,
    metavar="FILE_PATH",
    help=config_option_help,
)
@click.option(
    "--dir",
    "directory",
    default=None,
    metavar="PATH",
    help="Build fragment in directory. Default to current directory.",
)
@click.option(
    "--name",
    "project_name",
    default=None,
    help="Pass a custom project name.",
)
@click.option(
    "--version",
    "project_version",
    default=None,
    help="Render the news fragments using given version.",
)
@click.option(
    "--date",
    "project_date",
    default=None,
    help="Render the news fragments using the given date.",
)
@click.option(
    "--yes",
    "answer_yes",
    default=None,
    flag_value=True,
    help="Do not ask for confirmation to remove news fragments.",
    callback=_validate_answer,
)
@click.option(
    "--keep",
    "answer_keep",
    default=None,
    flag_value=True,
    help="Do not ask for confirmations. But keep news fragments.",
    callback=_validate_answer,
)
def _main(
    draft: bool,
    directory: str | None,
    config_file: str | None,
    project_name: str | None,
    project_version: str | None,
    project_date: str | None,
    answer_yes: bool,
    answer_keep: bool,
) -> None:
    """
    Build a combined news file from news fragment.
    """
    try:
        return __main(
            draft,
            directory,
            config_file,
            project_name,
            project_version,
            project_date,
            answer_yes,
            answer_keep,
        )
    except ConfigError as e:
        print(e, file=sys.stderr)
        sys.exit(1)


def __main(
    draft: bool,
    directory: str | None,
    config_file: str | None,
    project_name: str | None,
    project_version: str | None,
    project_date: str | None,
    answer_yes: bool,
    answer_keep: bool,
) -> None:
    """
    The main entry point.
    """
    base_directory, config = load_config_from_options(directory, config_file)
    to_err = draft

    if project_version is None:
        project_version = config.version
    if project_version is None:
        if not config.package:
            raise UsageError(
                "'--version' is required since the config file does "
                "not contain 'version' or 'package'."
            )
        project_version = get_version(
            os.path.join(base_directory, config.package_dir), config.package
        ).strip()

    click.echo("Loading template...", err=to_err)
    if isinstance(config.template, tuple):
        template = (
            resources.files(config.template[0])
            .joinpath(config.template[1])
            .read_text(encoding="utf-8")
        )
        template_extension = os.path.splitext(config.template[1])[1]
    else:
        template = Path(config.template).read_text(encoding="utf-8")
        template_extension = os.path.splitext(config.template)[1]
    is_markdown = template_extension.lower() == ".md"

    click.echo("Finding news fragments...", err=to_err)

    fragment_contents, fragment_files = find_fragments(
        base_directory,
        config,
        # Fail if any fragment filenames are invalid only if ignore list is set
        # (this maintains backward compatibility):
        strict=(config.ignore is not None),
    )
    fragment_filenames = [filename for (filename, _category) in fragment_files]

    click.echo("Rendering news fragments...", err=to_err)
    fragments = split_fragments(
        fragment_contents, config.types, all_bullets=config.all_bullets
    )

    if project_name is None:
        project_name = config.name
        if not project_name:
            package = config.package
            if package:
                project_name = get_project_name(
                    os.path.abspath(os.path.join(base_directory, config.package_dir)),
                    package,
                )
            else:
                # Can't determine a project_name, but maybe it is not needed.
                project_name = ""

    if project_date is None:
        project_date = _get_date().strip()

    # Render the title in the template if the title format is set to "". It can
    # alternatively be set to False or a string, in either case it shouldn't be rendered
    # in the template.
    render_title = config.title_format == ""

    rendered = render_fragments(
        # The 0th underline is used for the top line
        template,
        config.issue_format,
        fragments,
        config.types,
        config.underlines[1:],
        config.wrap,
        {"name": project_name, "version": project_version, "date": project_date},
        top_underline=config.underlines[0],
        all_bullets=config.all_bullets,
        render_title=render_title,
    )

    if config.title_format:
        top_line = config.title_format.format(
            name=project_name, version=project_version, project_date=project_date
        )
        if is_markdown:
            parts = [f"# {top_line}"]
        else:
            parts = [top_line, config.underlines[0] * len(top_line)]
        parts.append(rendered)
        content = "\n".join(parts)
    else:
        top_line = ""
        content = rendered

    if draft:
        click.echo(
            "Draft only -- nothing has been written.\n"
            "What is seen below is what would be written.\n",
            err=to_err,
        )
        click.echo(content)
        return

    click.echo("Writing to newsfile...", err=to_err)
    news_file = config.filename

    if config.single_file is False:
        # The release notes for each version are stored in a separate file.
        # The name of that file is generated based on the current version and project.
        news_file = news_file.format(
            name=project_name, version=project_version, project_date=project_date
        )

    append_to_newsfile(
        base_directory,
        news_file,
        config.start_string,
        top_line,
        content,
        single_file=config.single_file,
    )

    click.echo("Staging newsfile...", err=to_err)
    _git.stage_newsfile(base_directory, news_file)

    if should_remove_fragment_files(
        fragment_filenames,
        answer_yes,
        answer_keep,
    ):
        click.echo("Removing news fragments...", err=to_err)
        _git.remove_files(fragment_filenames)

    click.echo("Done!", err=to_err)


def should_remove_fragment_files(
    fragment_filenames: list[str],
    answer_yes: bool,
    answer_keep: bool,
) -> bool:
    if not fragment_filenames:
        click.echo("No news fragments to remove. Skipping!")
        return False
    try:
        if answer_keep:
            click.echo("Keeping the following files:")
            # Not proceeding with the removal of the files.
            return False

        if answer_yes:
            click.echo("Removing the following files:")
        else:
            click.echo("I want to remove the following files:")
    finally:
        # Will always be printed, even for answer_keep to help with possible troubleshooting
        for filename in fragment_filenames:
            click.echo(filename)

    if answer_yes or click.confirm("Is it okay if I remove those files?", default=True):
        return True
    return False


if __name__ == "__main__":  # pragma: no cover
    _main()



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.


from __future__ import annotations

import os
import re
import textwrap

from collections import defaultdict
from pathlib import Path
from typing import Any, DefaultDict, Iterable, Iterator, Mapping, NamedTuple, Sequence

from click import ClickException
from jinja2 import Template

from towncrier._settings.load import Config


# Returns issue, category and counter or (None, None, None) if the basename
# could not be parsed or doesn't contain a valid category.
def parse_newfragment_basename(
    basename: str, frag_type_names: Iterable[str]
) -> tuple[str, str, int] | tuple[None, None, None]:
    invalid = (None, None, None)
    parts = basename.split(".")

    if len(parts) == 1:
        return invalid

    # There are at least 2 parts. Search for a valid category from the second
    # part onwards starting at the back.
    # The category is used as the reference point in the parts list to later
    # infer the issue number and counter value.
    for i in reversed(range(1, len(parts))):
        if parts[i] in frag_type_names:
            # Current part is a valid category according to given definitions.
            category = parts[i]
            # Use all previous parts as the issue number.
            # NOTE: This allows news fragment names like fix-1.2.3.feature or
            # something-cool.feature.ext for projects that don't use issue
            # numbers in news fragment names.
            issue = ".".join(parts[0:i]).strip()
            # If the issue is an integer, remove any leading zeros (to resolve
            # issue #126).
            if issue.isdigit():
                issue = str(int(issue))
            counter = 0
            # Use the following part as the counter if it exists and is a valid
            # digit.
            if len(parts) > (i + 1) and parts[i + 1].isdigit():
                counter = int(parts[i + 1])
            return issue, category, counter
    else:
        # No valid category found.
        return invalid


class FragmentsPath:
    """
    A helper to get the full path to a fragments directory.

    This is a callable that optionally takes a section directory and returns the full
    path to the fragments directory for that section (or the default if no section is
    provided).
    """

    def __init__(self, base_directory: str, config: Config):
        self.base_directory = base_directory
        self.config = config
        if config.directory is not None:
            self.base_directory = os.path.abspath(
                os.path.join(base_directory, config.directory)
            )
            self.append_directory = ""
        else:
            self.base_directory = os.path.abspath(
                os.path.join(base_directory, config.package_dir, config.package)
            )
            self.append_directory = "newsfragments"

    def __call__(self, section_directory: str = "") -> str:
        return os.path.join(
            self.base_directory, section_directory, self.append_directory
        )


# Returns a structure like:
#
# {
#     "": {
#         ("142", "misc", 1): "",
#         ("1", "feature", 1): "some cool description",
#     },
#     "Names": {},
#     "Web": {("3", "bugfix", 1): "Fixed a thing"},
# }
#
# and a list like:
# [
#    ("/path/to/fragments/142.misc.1", "misc"),
#    ("/path/to/fragments/1.feature.1", "feature"),
# ]
#
# We should really use attrs.
def find_fragments(
    base_directory: str,
    config: Config,
    strict: bool,
) -> tuple[Mapping[str, Mapping[tuple[str, str, int], str]], list[tuple[str, str]]]:
    """
    Sections are a dictonary of section names to paths.

    If strict, raise ClickException if any fragments have an invalid name.
    """
    ignored_files = {".gitignore", ".keep", "readme", "readme.md", "readme.rst"}
    if config.ignore:
        ignored_files.update(filename.lower() for filename in config.ignore)

    get_section_path = FragmentsPath(base_directory, config)

    content = {}
    fragment_files = []
    # Multiple orphan news fragments are allowed per section, so initialize a counter
    # that can be incremented automatically.
    orphan_fragment_counter: DefaultDict[str | None, int] = defaultdict(int)

    for key, section_dir in config.sections.items():
        section_dir = get_section_path(section_dir)

        try:
            files = os.listdir(section_dir)
        except FileNotFoundError:
            files = []

        file_content = {}

        for basename in files:
            if basename.lower() in ignored_files:
                continue

            issue, category, counter = parse_newfragment_basename(
                basename, config.types
            )
            if category is None:
                if strict and issue is None:
                    raise ClickException(
                        f"Invalid news fragment name: {basename}\n"
                        "If this filename is deliberate, add it to "
                        "'ignore' in your configuration."
                    )
                continue
            assert issue is not None
            assert counter is not None
            if config.orphan_prefix and issue.startswith(config.orphan_prefix):
                issue = ""
                # Use and increment the orphan news fragment counter.
                counter = orphan_fragment_counter[category]
                orphan_fragment_counter[category] += 1

            full_filename = os.path.join(section_dir, basename)
            fragment_files.append((full_filename, category))
            data = Path(full_filename).read_text(encoding="utf-8", errors="replace")

            if (issue, category, counter) in file_content:
                raise ValueError(
                    "multiple files for {}.{} in {}".format(
                        issue, category, section_dir
                    )
                )
            file_content[issue, category, counter] = data

        content[key] = file_content

    return content, fragment_files


def indent(text: str, prefix: str) -> str:
    """
    Adds `prefix` to the beginning of non-empty lines in `text`.
    """

    # Based on Python 3's textwrap.indent
    def prefixed_lines() -> Iterator[str]:
        for line in text.splitlines(True):
            yield (prefix + line if line.strip() else line)

    return "".join(prefixed_lines())


# Takes the output from find_fragments above. Probably it would be useful to
# add an example output here. Next time someone digs deep enough to figure it
# out, please do so...
def split_fragments(
    fragments: Mapping[str, Mapping[tuple[str, str, int], str]],
    definitions: Mapping[str, Mapping[str, Any]],
    all_bullets: bool = True,
) -> Mapping[str, Mapping[str, Mapping[str, Sequence[str]]]]:
    output = {}

    for section_name, section_fragments in fragments.items():
        section: dict[str, dict[str, list[str]]] = {}

        for (issue, category, counter), content in section_fragments.items():
            if all_bullets:
                # By default all fragmetns are append by "-" automatically,
                # and need to be indented because of that.
                # (otherwise, assume they are formatted correctly)
                content = indent(content.strip(), "  ")[2:]
            else:
                # Assume the text is formatted correctly
                content = content.rstrip()

            if definitions[category]["showcontent"] is False and issue:
                # If this category is not supposed to show content (and we have an
                # issue) then we should just add the issue to the section rather than
                # the content. If there isn't an issue, still add the content so that
                # it's recorded.
                content = ""

            texts = section.setdefault(category, {})

            issues = texts.setdefault(content, [])
            if issue:
                # Only add the issue if we have one (it can be blank for orphan news
                # fragments).
                issues.append(issue)
                issues.sort()

        output[section_name] = section

    return output


class IssueParts(NamedTuple):
    is_digit: bool
    has_digit: bool
    non_digit_part: str
    number: int


def issue_key(issue: str) -> IssueParts:
    """
    Used to sort the issue ID inside a news fragment in a human-friendly way.

    Issue IDs are grouped by their non-integer part, then sorted by their integer part.

    For backwards compatible consistency, issues without no number are sorted first and
    digit only issues are sorted last.

    For example::

    >>> sorted(["2", "#11", "#3", "gh-10", "gh-4", "omega", "alpha"], key=issue_key)
    ['alpha', 'omega', '#3', '#11', 'gh-4', 'gh-10', '2']
    """
    if issue.isdigit():
        return IssueParts(
            is_digit=True, has_digit=True, non_digit_part="", number=int(issue)
        )
    match = re.search(r"\d+", issue)
    if not match:
        return IssueParts(
            is_digit=False, has_digit=False, non_digit_part=issue, number=-1
        )
    return IssueParts(
        is_digit=False,
        has_digit=True,
        non_digit_part=issue[: match.start()] + issue[match.end() :],
        number=int(match.group()),
    )


def entry_key(entry: tuple[str, Sequence[str]]) -> tuple[str, list[IssueParts]]:
    content, issues = entry
    # Orphan news fragments (those without any issues) should sort last by content.
    return "" if issues else content, [issue_key(issue) for issue in issues]


def bullet_key(entry: tuple[str, Sequence[str]]) -> int:
    text, _ = entry
    if not text:
        return -1
    if text[:2] == "- ":
        return 0
    elif text[:2] == "* ":
        return 1
    elif text[:3] == "#. ":
        return 2
    return 3


def render_issue(issue_format: str | None, issue: str) -> str:
    if issue_format is None:
        try:
            int(issue)
            return "#" + issue
        except Exception:
            return issue
    else:
        return issue_format.format(issue=issue)


def render_fragments(
    template: str,
    issue_format: str | None,
    fragments: Mapping[str, Mapping[str, Mapping[str, Sequence[str]]]],
    definitions: Mapping[str, Mapping[str, Any]],
    underlines: Sequence[str],
    wrap: bool,
    versiondata: Mapping[str, str],
    top_underline: str = "=",
    all_bullets: bool = False,
    render_title: bool = True,
) -> str:
    """
    Render the fragments into a news file.
    """

    jinja_template = Template(template, trim_blocks=True)

    data: dict[str, dict[str, dict[str, list[str]]]] = {}
    issues_by_category: dict[str, dict[str, list[str]]] = {}

    for section_name, section_value in fragments.items():
        data[section_name] = {}
        issues_by_category[section_name] = {}

        for category_name, category_value in section_value.items():
            category_issues: set[str] = set()
            # Suppose we start with an ordering like this:
            #
            # - Fix the thing (#7, #123, #2)
            # - Fix the other thing (#1)

            # First we sort the issues inside each line:
            #
            # - Fix the thing (#2, #7, #123)
            # - Fix the other thing (#1)
            entries = []
            for text, issues in category_value.items():
                entries.append((text, sorted(issues, key=issue_key)))
                category_issues.update(issues)

            # Then we sort the lines:
            #
            # - Fix the other thing (#1)
            # - Fix the thing (#2, #7, #123)
            entries.sort(key=entry_key)
            if not all_bullets:
                entries.sort(key=bullet_key)

            # Then we put these nicely sorted entries back in an ordered dict
            # for the template, after formatting each issue number
            categories = {}
            for text, issues in entries:
                rendered = [render_issue(issue_format, i) for i in issues]
                categories[text] = rendered

            data[section_name][category_name] = categories
            issues_by_category[section_name][category_name] = [
                render_issue(issue_format, i)
                for i in sorted(category_issues, key=issue_key)
            ]

    done = []

    def get_indent(text: str) -> str:
        # If bullets are not assumed and we wrap, the subsequent
        # indentation depends on whether or not this is a bullet point.
        # (it is probably usually best to disable wrapping in that case)
        if all_bullets or text[:2] == "- " or text[:2] == "* ":
            return "  "
        elif text[:3] == "#. ":
            return "   "
        return ""

    res = jinja_template.render(
        render_title=render_title,
        sections=data,
        definitions=definitions,
        underlines=underlines,
        versiondata=versiondata,
        top_underline=top_underline,
        get_indent=get_indent,  # simplify indentation in the jinja template.
        issues_by_category=issues_by_category,
    )

    for line in res.split("\n"):
        if wrap:
            done.append(
                textwrap.fill(
                    line,
                    width=79,
                    subsequent_indent=get_indent(line),
                    break_long_words=False,
                    break_on_hyphens=False,
                )
            )
        else:
            done.append(line)

    return "\n".join(done)



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.


from twisted.trial.unittest import TestCase

from towncrier import _git


class TestGit(TestCase):
    def test_empty_remove(self):
        """
        If remove_files gets an empty list, it returns gracefully.
        """
        _git.remove_files([])



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

"""
Responsible for writing the built news fragments to a news file without
affecting existing content.
"""

from __future__ import annotations

import sys

from pathlib import Path
from typing import Any


if sys.version_info < (3, 10):
    # Compatibility shim for newline parameter to write_text, added in 3.10
    def _newline_write_text(path: Path, content: str, **kwargs: Any) -> None:
        with path.open("w", **kwargs) as strm:  # pragma: no branch
            strm.write(content)

else:

    def _newline_write_text(path: Path, content: str, **kwargs: Any) -> None:
        path.write_text(content, **kwargs)


def append_to_newsfile(
    directory: str,
    filename: str,
    start_string: str,
    top_line: str,
    content: str,
    single_file: bool,
) -> None:
    """
    Write *content* to *directory*/*filename* behind *start_string*.

    Double-check *top_line* (i.e. the release header) is not already in the
    file.

    if *single_file* is True, add it to an existing file, otherwise create a
    fresh one.
    """
    news_file = Path(directory) / filename

    header, prev_body = _figure_out_existing_content(
        news_file, start_string, single_file
    )

    if top_line and top_line in prev_body:
        raise ValueError("It seems you've already produced newsfiles for this version?")

    _newline_write_text(
        news_file,
        # If there is no previous body that means we're writing a brand new news file.
        # We don't want extra whitespace at the end of this new file.
        header + (content + prev_body if prev_body else content.rstrip() + "\n"),
        encoding="utf-8",
        # Leave newlines alone. This probably leads to inconsistent newlines,
        # because we've loaded existing content with universal newlines, but that's
        # the original behavior.
        newline="",
    )


def _figure_out_existing_content(
    news_file: Path, start_string: str, single_file: bool
) -> tuple[str, str]:
    """
    Try to read *news_file* and split it into header (everything before
    *start_string*) and the old body (everything after *start_string*).

    If there's no *start_string*, return empty header.

    Empty file and per-release files have neither.
    """
    if not single_file or not news_file.exists():
        # Per-release news files always start empty.
        # Non-existent files have no existing content.
        return "", ""

    content = Path(news_file).read_text(encoding="utf-8")

    t = content.split(start_string, 1)
    if len(t) == 2:
        return f"{t[0].rstrip()}\n\n{start_string}\n", t[1].lstrip()

    return "", content.lstrip()



"""Subpackage to handle settings parsing."""

from __future__ import annotations

from towncrier._settings import load


load_config = load.load_config
ConfigError = load.ConfigError
load_config_from_options = load.load_config_from_options

# Help message for --config CLI option, shared by all sub-commands.
config_option_help = (
    "Pass a custom config file at FILE_PATH. "
    "Default: towncrier.toml or pyproject.toml file, "
    "if both files exist, the first will take precedence."
)

__all__ = [
    "config_option_help",
    "load_config",
    "ConfigError",
    "load_config_from_options",
]



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

import os

from click.testing import CliRunner
from twisted.trial.unittest import TestCase

from .._settings import ConfigError, load_config
from .._shell import cli
from .helpers import with_isolated_runner, write


class TomlSettingsTests(TestCase):
    def mktemp_project(
        self, *, pyproject_toml: str = "", towncrier_toml: str = ""
    ) -> str:
        """
        Create a temporary directory with a pyproject.toml file in it.
        """
        project_dir = self.mktemp()
        os.makedirs(project_dir)

        if pyproject_toml:
            write(
                os.path.join(project_dir, "pyproject.toml"),
                pyproject_toml,
                dedent=True,
            )

        if towncrier_toml:
            write(
                os.path.join(project_dir, "towncrier.toml"),
                towncrier_toml,
                dedent=True,
            )

        return project_dir

    def test_base(self):
        """
        Test a "base config".
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                package = "foobar"
                orphan_prefix = "~"
            """
        )

        config = load_config(project_dir)
        self.assertEqual(config.package, "foobar")
        self.assertEqual(config.package_dir, ".")
        self.assertEqual(config.filename, "NEWS.rst")
        self.assertEqual(config.underlines, ("=", "-", "~"))
        self.assertEqual(config.orphan_prefix, "~")

    def test_markdown(self):
        """
        If the filename references an .md file and the builtin template doesn't have an
        extension, add .md rather than .rst.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                package = "foobar"
                filename = "NEWS.md"
            """
        )

        config = load_config(project_dir)

        self.assertEqual(config.filename, "NEWS.md")

        self.assertEqual(config.template, ("towncrier.templates", "default.md"))

    def test_explicit_template_extension(self):
        """
        If the filename references an .md file and the builtin template has an
        extension, don't change it.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                package = "foobar"
                filename = "NEWS.md"
                template = "towncrier:default.rst"
            """
        )

        config = load_config(project_dir)

        self.assertEqual(config.filename, "NEWS.md")
        self.assertEqual(config.template, ("towncrier.templates", "default.rst"))

    def test_template_extended(self):
        """
        The template can be any package and resource, and although we look for a
        resource's 'templates' package, it could also be in the specified resource
        directly.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                package = "foobar"
                template = "towncrier.templates:default.rst"
            """
        )

        config = load_config(project_dir)

        self.assertEqual(config.template, ("towncrier.templates", "default.rst"))

    def test_missing(self):
        """
        If the config file doesn't have the correct toml key, we error.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [something.else]
                blah='baz'
            """
        )

        with self.assertRaises(ConfigError) as e:
            load_config(project_dir)

        self.assertEqual(e.exception.failing_option, "all")

    def test_incorrect_single_file(self):
        """
        single_file must be a bool.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                single_file = "a"
            """
        )

        with self.assertRaises(ConfigError) as e:
            load_config(project_dir)

        self.assertEqual(e.exception.failing_option, "single_file")

    def test_incorrect_all_bullets(self):
        """
        all_bullets must be a bool.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                all_bullets = "a"
            """
        )

        with self.assertRaises(ConfigError) as e:
            load_config(project_dir)

        self.assertEqual(e.exception.failing_option, "all_bullets")

    def test_mistype_singlefile(self):
        """
        singlefile is not accepted, single_file is.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                singlefile = "a"
            """
        )

        with self.assertRaises(ConfigError) as e:
            load_config(project_dir)

        self.assertEqual(e.exception.failing_option, "singlefile")

    def test_towncrier_toml_preferred(self):
        """
        Towncrier prefers the towncrier.toml for autodetect over pyproject.toml.
        """
        project_dir = self.mktemp_project(
            towncrier_toml="""
                [tool.towncrier]
                package = "a"
            """,
            pyproject_toml="""
                [tool.towncrier]
                package = "b"
            """,
        )

        config = load_config(project_dir)
        self.assertEqual(config.package, "a")

    @with_isolated_runner
    def test_load_no_config(self, runner: CliRunner):
        """
        Calling the root CLI without an existing configuration file in the base directory,
        will exit with code 1 and an informative message is sent to standard output.
        """
        temp = self.mktemp()
        os.makedirs(temp)

        result = runner.invoke(cli, ("--dir", temp))

        self.assertEqual(
            result.output,
            f"No configuration file found.\nLooked back from: {os.path.abspath(temp)}\n",
        )
        self.assertEqual(result.exit_code, 1)

    @with_isolated_runner
    def test_load_explicit_missing_config(self, runner: CliRunner):
        """
        Calling the CLI with an incorrect explicit configuration file will exit with
        code 1 and an informative message is sent to standard output.
        """
        config = "not-there.toml"
        result = runner.invoke(cli, ("--config", config))

        self.assertEqual(result.exit_code, 1)
        self.assertEqual(
            result.output,
            f"Configuration file '{os.path.abspath(config)}' not found.\n",
        )

    def test_missing_template(self):
        """
        Towncrier will raise an exception saying when it can't find a template.
        """
        project_dir = self.mktemp_project(
            towncrier_toml="""
                [tool.towncrier]
                template = "foo.rst"
            """
        )

        with self.assertRaises(ConfigError) as e:
            load_config(project_dir)

        self.assertEqual(
            str(e.exception),
            "The template file '{}' does not exist.".format(
                os.path.normpath(os.path.join(project_dir, "foo.rst")),
            ),
        )

    def test_missing_template_in_towncrier(self):
        """
        Towncrier will raise an exception saying when it can't find a template
        from the Towncrier templates.
        """
        project_dir = self.mktemp_project(
            towncrier_toml="""
                [tool.towncrier]
                template = "towncrier:foo"
            """
        )

        with self.assertRaises(ConfigError) as e:
            load_config(project_dir)

        self.assertEqual(
            str(e.exception), "'towncrier' does not have a template named 'foo.rst'."
        )

    def test_custom_types_as_tables_array_deprecated(self):
        """
        Custom fragment categories can be defined inside
        the toml config file using an array of tables
        (a table name in double brackets).

        This functionality is considered deprecated, but we continue
        to support it to keep backward compatibility.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                package = "foobar"
                [[tool.towncrier.type]]
                directory="foo"
                name="Foo"
                showcontent=false

                [[tool.towncrier.type]]
                directory="spam"
                name="Spam"
                showcontent=true

                [[tool.towncrier.type]]
                directory="auto"
                name="Automatic"
                showcontent=true
                check=false
            """
        )
        config = load_config(project_dir)
        expected = [
            (
                "foo",
                {
                    "name": "Foo",
                    "showcontent": False,
                    "check": True,
                },
            ),
            (
                "spam",
                {
                    "name": "Spam",
                    "showcontent": True,
                    "check": True,
                },
            ),
            (
                "auto",
                {
                    "name": "Automatic",
                    "showcontent": True,
                    "check": False,
                },
            ),
        ]
        expected = dict(expected)
        actual = config.types
        self.assertDictEqual(expected, actual)

    def test_custom_types_as_tables(self):
        """
        Custom fragment categories can be defined inside
        the toml config file using tables.
        """
        project_dir = self.mktemp_project(
            pyproject_toml="""
                [tool.towncrier]
                package = "foobar"
                [tool.towncrier.fragment.feat]
                ignored_field="Bazz"
                [tool.towncrier.fragment.fix]
                [tool.towncrier.fragment.chore]
                name = "Other Tasks"
                showcontent = false
                [tool.towncrier.fragment.auto]
                name = "Automatic"
                check = false
            """
        )
        config = load_config(project_dir)
        expected = {
            "chore": {
                "name": "Other Tasks",
                "showcontent": False,
                "check": True,
            },
            "feat": {
                "name": "Feat",
                "showcontent": True,
                "check": True,
            },
            "fix": {
                "name": "Fix",
                "showcontent": True,
                "check": True,
            },
            "auto": {
                "name": "Automatic",
                "showcontent": True,
                "check": False,
            },
        }
        actual = config.types
        self.assertDictEqual(expected, actual)



# Copyright (c) Amber Brown, 2018
# See LICENSE for details.


from __future__ import annotations

import os
import sys

from subprocess import CalledProcessError
from typing import Container
from warnings import warn

import click

from ._builder import find_fragments
from ._git import get_remote_branches, list_changed_files_compared_to_branch
from ._settings import config_option_help, load_config_from_options


def _get_default_compare_branch(branches: Container[str]) -> str | None:
    if "origin/main" in branches:
        return "origin/main"
    if "origin/master" in branches:
        warn(
            'Using "origin/master" as default compare branch is deprecated '
            "and will be removed in a future version.",
            DeprecationWarning,
            stacklevel=2,
        )
        return "origin/master"
    return None


@click.command(name="check")
@click.option(
    "--compare-with",
    default=None,
    metavar="BRANCH",
    help=(
        "Checks files changed running git diff --name-only BRANCH... "
        "BRANCH is the branch to be compared with. "
        "Default to origin/main"
    ),
)
@click.option(
    "--dir",
    "directory",
    default=None,
    metavar="PATH",
    help="Check fragment in directory. Default to current directory.",
)
@click.option(
    "--config",
    "config",
    default=None,
    metavar="FILE_PATH",
    help=config_option_help,
)
def _main(compare_with: str | None, directory: str | None, config: str | None) -> None:
    """
    Check for new fragments on a branch.
    """
    __main(compare_with, directory, config)


def __main(
    comparewith: str | None, directory: str | None, config_path: str | None
) -> None:
    base_directory, config = load_config_from_options(directory, config_path)

    if comparewith is None:
        comparewith = _get_default_compare_branch(
            get_remote_branches(base_directory=base_directory)
        )

    if comparewith is None:
        click.echo("Could not detect default branch. Aborting.")
        sys.exit(1)

    try:
        files_changed = list_changed_files_compared_to_branch(
            base_directory, comparewith
        )
    except CalledProcessError as e:
        click.echo("git produced output while failing:")
        click.echo(e.output)
        raise

    if not files_changed:
        click.echo(
            f"On {comparewith} branch, or no diffs, so no newsfragment required."
        )
        sys.exit(0)

    files = {os.path.abspath(path) for path in files_changed}

    click.echo("Looking at these files:")
    click.echo("----")
    for n, change in enumerate(files, start=1):
        click.echo(f"{n}. {change}")
    click.echo("----")

    # This will fail if any fragment files have an invalid name:
    all_fragment_files = find_fragments(base_directory, config, strict=True)[1]

    news_file = os.path.normpath(os.path.join(base_directory, config.filename))
    if news_file in files:
        click.echo("Checks SKIPPED: news file changes detected.")
        sys.exit(0)

    fragments = set()  # will only include fragments of types that are checked
    unchecked_fragments = set()  # will include fragments of types that are not checked
    for fragment_filename, category in all_fragment_files:
        if config.types[category]["check"]:
            fragments.add(fragment_filename)
        else:
            unchecked_fragments.add(fragment_filename)
    fragments_in_branch = fragments & files

    if not fragments_in_branch:
        unchecked_fragments_in_branch = unchecked_fragments & files
        if unchecked_fragments:
            click.echo("Found newsfragments of unchecked types in the branch:")
            for n, fragment in enumerate(unchecked_fragments_in_branch, start=1):
                click.echo(f"{n}. {fragment}")
        else:
            click.echo("No new newsfragments found on this branch.")
        sys.exit(1)
    else:
        click.echo("Found:")
        for n, fragment in enumerate(fragments_in_branch, start=1):
            click.echo(f"{n}. {fragment}")
        sys.exit(0)


if __name__ == "__main__":  # pragma: no cover
    _main()



"""
Provides towncrier version information.
"""

# For dev   - 23.11.0.dev0
# For RC    - 23.11.0rc1  (release candidate starts at 1)
# For final - 23.11.0
# make sure to follow PEP440
__version__ = "24.7.0"

_hatchling_version = __version__
__all__ = ["_hatchling_version"]



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

import os
import tempfile

from datetime import date
from pathlib import Path
from subprocess import call
from textwrap import dedent
from unittest.mock import patch

from click.testing import CliRunner
from twisted.trial.unittest import TestCase

from .._shell import cli
from ..build import _main
from .helpers import read, with_git_project, with_project, write


class TestCli(TestCase):
    maxDiff = None

    @with_project()
    def _test_command(self, command, runner):
        # Off the shelf newsfragment
        with open("foo/newsfragments/123.feature", "w") as f:
            f.write("Adds levitation")
        # Towncrier treats this as 124.feature, ignoring .rst extension
        with open("foo/newsfragments/124.feature.rst", "w") as f:
            f.write("Extends levitation")
        # Towncrier supports non-numeric newsfragment names.
        with open("foo/newsfragments/baz.feature.rst", "w") as f:
            f.write("Baz levitation")
        # Towncrier supports files that have a dot in the name of the
        # newsfragment
        with open("foo/newsfragments/fix-1.2.feature", "w") as f:
            f.write("Baz fix levitation")
        # Towncrier supports fragments not linked to a feature
        with open("foo/newsfragments/+anything.feature", "w") as f:
            f.write("Orphaned feature")
        with open("foo/newsfragments/+xxx.feature", "w") as f:
            f.write("Another orphaned feature")
        with open("foo/newsfragments/+123_orphaned.feature", "w") as f:
            f.write("An orphaned feature starting with a number")
        with open("foo/newsfragments/+12.3_orphaned.feature", "w") as f:
            f.write("An orphaned feature starting with a dotted number")
        with open("foo/newsfragments/+orphaned_123.feature", "w") as f:
            f.write("An orphaned feature ending with a number")
        with open("foo/newsfragments/+orphaned_12.3.feature", "w") as f:
            f.write("An orphaned feature ending with a dotted number")
        # Towncrier ignores files that don't have a dot
        with open("foo/newsfragments/README", "w") as f:
            f.write("Blah blah")
        # And files that don't have a valid category
        with open("foo/newsfragments/README.rst", "w") as f:
            f.write("**Blah blah**")

        result = runner.invoke(command, ["--draft", "--date", "01-01-2001"])

        self.assertEqual(0, result.exit_code, result.output)
        self.assertEqual(
            result.output,
            dedent(
                """\
                Loading template...
                Finding news fragments...
                Rendering news fragments...
                Draft only -- nothing has been written.
                What is seen below is what would be written.

                Foo 1.2.3 (01-01-2001)
                ======================

                Features
                --------

                - Baz levitation (baz)
                - Baz fix levitation (fix-1.2)
                - Adds levitation (#123)
                - Extends levitation (#124)
                - An orphaned feature ending with a dotted number
                - An orphaned feature ending with a number
                - An orphaned feature starting with a dotted number
                - An orphaned feature starting with a number
                - Another orphaned feature
                - Orphaned feature



                """
            ),
        )

    def test_command(self):
        self._test_command(cli)

    def test_subcommand(self):
        self._test_command(_main)

    @with_project()
    def test_in_different_dir_dir_option(self, runner):
        """
        The current working directory doesn't matter as long as we pass
        the correct one.
        """
        project_dir = Path(".").resolve()

        Path("foo/newsfragments/123.feature").write_text("Adds levitation")
        # Ensure our assetion below is meaningful.
        self.assertFalse((project_dir / "NEWS.rst").exists())

        # Create a temporary directory, run Towncrier from there and assert
        # it didn't litter into it.
        td = tempfile.TemporaryDirectory()
        self.addCleanup(td.cleanup)

        os.chdir(td.name)
        result = runner.invoke(cli, ("--yes", "--dir", str(project_dir)))

        self.assertEqual([], list(Path(td.name).glob("*")))
        self.assertEqual(0, result.exit_code)
        self.assertTrue((project_dir / "NEWS.rst").exists())

    @with_project()
    def test_traverse_up_to_find_config(self, runner):
        """
        When the current directory doesn't contain the configuration file, Towncrier
        will traverse up the directory tree until it finds it.
        """
        os.chdir("foo")
        result = runner.invoke(_main, ["--draft", "--date", "01-01-2001"])
        self.assertEqual(0, result.exit_code, result.output)

    @with_project()
    def test_in_different_dir_config_option(self, runner):
        """
        The current working directory and the location of the configuration
        don't matter as long as we pass correct paths to the directory and the
        config file.
        """
        project_dir = Path(".").resolve()

        Path("foo/newsfragments/123.feature").write_text("Adds levitation")
        # Ensure our assetion below is meaningful.
        self.assertFalse((project_dir / "NEWS.rst").exists())

        # Create a temporary directory, move the config file there, run
        # Towncrier from there, and assert it didn't litter into it.
        td = tempfile.TemporaryDirectory()
        self.addCleanup(td.cleanup)

        os.chdir(td.name)
        (project_dir / "pyproject.toml").rename("pyproject.toml")
        result = runner.invoke(
            cli, ("--yes", "--config", "pyproject.toml", "--dir", str(project_dir))
        )

        # There's only pyproject.toml in this directory.
        self.assertEqual(
            [Path(td.name) / "pyproject.toml"], list(Path(td.name).glob("*"))
        )
        self.assertEqual(0, result.exit_code)
        self.assertTrue((project_dir / "NEWS.rst").exists())

    @with_project(
        config="""
        [tool.towncrier]
        directory = "changelog.d"
        """
    )
    def test_in_different_dir_with_nondefault_newsfragments_directory(self, runner):
        """
        Using the `--dir` CLI argument, the NEWS file can
        be generated in a sub-directory from fragments
        that are relatives to that sub-directory.

        The path passed to `--dir` becomes the
        working directory.
        """
        Path("foo/foo").mkdir(parents=True)
        Path("foo/foo/__init__.py").write_text("")
        Path("foo/changelog.d").mkdir()
        Path("foo/changelog.d/123.feature").write_text("Adds levitation")
        self.assertFalse(Path("foo/NEWS.rst").exists())

        result = runner.invoke(
            cli,
            ("--yes", "--config", "pyproject.toml", "--dir", "foo", "--version", "1.0"),
        )

        self.assertEqual(0, result.exit_code)
        self.assertTrue(Path("foo/NEWS.rst").exists())

    @with_project()
    def test_no_newsfragment_directory(self, runner):
        """
        A missing newsfragment directory acts as if there are no changes.
        """
        os.rmdir("foo/newsfragments")

        result = runner.invoke(_main, ["--draft", "--date", "01-01-2001"])

        self.assertEqual(0, result.exit_code)
        self.assertIn("No significant changes.\n", result.output)

    @with_project()
    def test_no_newsfragments_draft(self, runner):
        """
        An empty newsfragment directory acts as if there are no changes.
        """
        result = runner.invoke(_main, ["--draft", "--date", "01-01-2001"])

        self.assertEqual(0, result.exit_code)
        self.assertIn("No significant changes.\n", result.output)

    @with_project()
    def test_no_newsfragments(self, runner):
        """
        An empty newsfragment directory acts as if there are no changes and
        removing files handles it gracefully.
        """
        result = runner.invoke(_main, ["--date", "01-01-2001"])

        news = read("NEWS.rst")

        self.assertEqual(0, result.exit_code)
        self.assertIn("No significant changes.\n", news)

    @with_project()
    def test_collision(self, runner):
        # Note that both are 123.feature
        with open("foo/newsfragments/123.feature", "w") as f:
            f.write("Adds levitation")
        with open("foo/newsfragments/123.feature.rst", "w") as f:
            f.write("Extends levitation")

        result = runner.invoke(_main, ["--draft", "--date", "01-01-2001"])

        # This should fail
        self.assertEqual(type(result.exception), ValueError)
        self.assertIn("multiple files for 123.feature", str(result.exception))

    def test_section_and_type_sorting(self):
        """
        Sections and types should be output in the same order that they're
        defined in the config file.
        """

        runner = CliRunner()

        def run_order_scenario(sections, types):
            with runner.isolated_filesystem():
                with open("pyproject.toml", "w") as f:
                    f.write(
                        dedent(
                            """
                    [tool.towncrier]
                        package = "foo"
                        directory = "news"

                    """
                        )
                    )

                    for section in sections:
                        f.write(
                            dedent(
                                """
                        [[tool.towncrier.section]]
                            path = "{section}"
                            name = "{section}"
                        """.format(
                                    section=section
                                )
                            )
                        )

                    for type_ in types:
                        f.write(
                            dedent(
                                """
                        [[tool.towncrier.type]]
                            directory = "{type_}"
                            name = "{type_}"
                            showcontent = true
                        """.format(
                                    type_=type_
                                )
                            )
                        )

                os.mkdir("foo")
                with open("foo/__init__.py", "w") as f:
                    f.write('__version__ = "1.2.3"\n')
                os.mkdir("news")
                for section in sections:
                    sectdir = "news/" + section
                    os.mkdir(sectdir)
                    for type_ in types:
                        with open(f"{sectdir}/1.{type_}", "w") as f:
                            f.write(f"{section} {type_}")

                return runner.invoke(
                    _main, ["--draft", "--date", "01-01-2001"], catch_exceptions=False
                )

        result = run_order_scenario(["section-a", "section-b"], ["type-1", "type-2"])
        self.assertEqual(0, result.exit_code)
        self.assertEqual(
            result.output,
            "Loading template...\nFinding news fragments...\nRendering news "
            "fragments...\nDraft only -- nothing has been written.\nWhat is "
            "seen below is what would be written.\n\nFoo 1.2.3 (01-01-2001)"
            "\n======================"
            + dedent(
                """
                  section-a
                  ---------

                  type-1
                  ~~~~~~

                  - section-a type-1 (#1)


                  type-2
                  ~~~~~~

                  - section-a type-2 (#1)


                  section-b
                  ---------

                  type-1
                  ~~~~~~

                  - section-b type-1 (#1)


                  type-2
                  ~~~~~~

                  - section-b type-2 (#1)



            """
            ),
        )

        result = run_order_scenario(["section-b", "section-a"], ["type-2", "type-1"])
        self.assertEqual(0, result.exit_code)
        self.assertEqual(
            result.output,
            "Loading template...\nFinding news fragments...\nRendering news "
            "fragments...\nDraft only -- nothing has been written.\nWhat is "
            "seen below is what would be written.\n\nFoo 1.2.3 (01-01-2001)"
            "\n======================"
            + dedent(
                """
                  section-b
                  ---------

                  type-2
                  ~~~~~~

                  - section-b type-2 (#1)


                  type-1
                  ~~~~~~

                  - section-b type-1 (#1)


                  section-a
                  ---------

                  type-2
                  ~~~~~~

                  - section-a type-2 (#1)


                  type-1
                  ~~~~~~

                  - section-a type-1 (#1)



            """
            ),
        )

    @with_git_project()
    def test_draft_no_date(self, runner, commit):
        """
        If no date is passed, today's date is used.
        """
        fragment_path1 = "foo/newsfragments/123.feature"
        fragment_path2 = "foo/newsfragments/124.feature.rst"
        with open(fragment_path1, "w") as f:
            f.write("Adds levitation")
        with open(fragment_path2, "w") as f:
            f.write("Extends levitation")

        commit()

        today = date.today()
        result = runner.invoke(_main, ["--draft"])

        self.assertEqual(0, result.exit_code)
        self.assertIn(f"Foo 1.2.3 ({today.isoformat()})", result.output)

    @with_git_project()
    def test_no_confirmation(self, runner, commit):
        fragment_path1 = "foo/newsfragments/123.feature"
        fragment_path2 = "foo/newsfragments/124.feature.rst"
        with open(fragment_path1, "w") as f:
            f.write("Adds levitation")
        with open(fragment_path2, "w") as f:
            f.write("Extends levitation")

        commit()

        result = runner.invoke(_main, ["--date", "01-01-2001", "--yes"])

        self.assertEqual(0, result.exit_code)
        path = "NEWS.rst"
        self.assertTrue(os.path.isfile(path))
        self.assertFalse(os.path.isfile(fragment_path1))
        self.assertFalse(os.path.isfile(fragment_path2))

    @with_git_project()
    def test_keep_fragments(self, runner, commit):
        """
        The `--keep` option will build the full final news file
        without deleting the fragment files and without
        any extra CLI interaction or confirmation.
        """
        fragment_path1 = "foo/newsfragments/123.feature"
        fragment_path2 = "foo/newsfragments/124.feature.rst"
        with open(fragment_path1, "w") as f:
            f.write("Adds levitation")
        with open(fragment_path2, "w") as f:
            f.write("Extends levitation")

        commit()

        result = runner.invoke(_main, ["--date", "01-01-2001", "--keep"])

        self.assertEqual(0, result.exit_code)
        # The NEWS file is created.
        # So this is not just `--draft`.
        self.assertTrue(os.path.isfile("NEWS.rst"))
        self.assertTrue(os.path.isfile(fragment_path1))
        self.assertTrue(os.path.isfile(fragment_path2))

    @with_git_project()
    def test_yes_keep_error(self, runner, commit):
        """
        It will fail to perform any action when the
        conflicting --keep and --yes options are provided.

        Called twice with the different order of --keep and --yes options
        to make sure both orders are validated since click triggers the validator
        in the order it parses the command line.
        """
        fragment_path1 = "foo/newsfragments/123.feature"
        fragment_path2 = "foo/newsfragments/124.feature.rst"
        with open(fragment_path1, "w") as f:
            f.write("Adds levitation")
        with open(fragment_path2, "w") as f:
            f.write("Extends levitation")

        commit()

        result = runner.invoke(_main, ["--date", "01-01-2001", "--yes", "--keep"])
        self.assertEqual(1, result.exit_code)

        result = runner.invoke(_main, ["--date", "01-01-2001", "--keep", "--yes"])
        self.assertEqual(1, result.exit_code)

    @with_git_project()
    def test_confirmation_says_no(self, runner, commit):
        """
        If the user says "no" to removing the newsfragements, we end up with
        a NEWS.rst AND the newsfragments.
        """
        fragment_path1 = "foo/newsfragments/123.feature"
        fragment_path2 = "foo/newsfragments/124.feature.rst"
        with open(fragment_path1, "w") as f:
            f.write("Adds levitation")
        with open(fragment_path2, "w") as f:
            f.write("Extends levitation")

        commit()

        with patch("towncrier.build.click.confirm") as m:
            m.return_value = False
            result = runner.invoke(_main, [])

        self.assertEqual(0, result.exit_code)
        path = "NEWS.rst"
        self.assertTrue(os.path.isfile(path))
        self.assertTrue(os.path.isfile(fragment_path1))
        self.assertTrue(os.path.isfile(fragment_path2))

    def test_needs_config(self):
        """
        Towncrier needs a configuration file.
        """
        runner = CliRunner()

        with runner.isolated_filesystem():
            result = runner.invoke(_main, ["--draft"])

        self.assertEqual(1, result.exit_code, result.output)
        self.assertTrue(result.output.startswith("No configuration file found."))

    @with_project(config="[tool.towncrier]")
    def test_needs_version(self, runner: CliRunner):
        """
        If the configuration file doesn't specify a version or a package, the version
        option is required.
        """
        result = runner.invoke(_main, ["--draft"], catch_exceptions=False)

        self.assertEqual(2, result.exit_code)
        self.assertIn("Error: '--version' is required", result.output)

    @with_project()
    def test_projectless_changelog(self, runner):
        """In which a directory containing news files is built into a changelog

        - without a Python project or version number. We override the
        project title from the commandline.
        """
        # Remove the version from the project
        Path("foo/__init__.py").unlink()

        with open("foo/newsfragments/123.feature", "w") as f:
            f.write("Adds levitation")
        # Towncrier ignores .rst extension
        with open("foo/newsfragments/124.feature.rst", "w") as f:
            f.write("Extends levitation")

        result = runner.invoke(
            _main,
            [
                "--name",
                "FooBarBaz",
                "--version",
                "7.8.9",
                "--date",
                "01-01-2001",
                "--draft",
            ],
        )

        self.assertEqual(0, result.exit_code)
        self.assertEqual(
            result.output,
            dedent(
                """
            Loading template...
            Finding news fragments...
            Rendering news fragments...
            Draft only -- nothing has been written.
            What is seen below is what would be written.

            FooBarBaz 7.8.9 (01-01-2001)
            ============================

            Features
            --------

            - Adds levitation (#123)
            - Extends levitation (#124)



            """
            ).lstrip(),
        )

    @with_project(
        config="""
        [tool.towncrier]
        version = "7.8.9"
        """
    )
    def test_version_in_config(self, runner):
        """Calling towncrier with version defined in configfile.

        Specifying a version in toml file will be helpful if version
        is maintained by i.e. bumpversion and it's not a python project.
        """
        os.mkdir("newsfragments")
        with open("newsfragments/123.feature", "w") as f:
            f.write("Adds levitation")

        result = runner.invoke(_main, ["--date", "01-01-2001", "--draft"])

        self.assertEqual(0, result.exit_code, result.output)
        self.assertEqual(
            result.output,
            dedent(
                """
            Loading template...
            Finding news fragments...
            Rendering news fragments...
            Draft only -- nothing has been written.
            What is seen below is what would be written.

            7.8.9 (01-01-2001)
            ==================

            Features
            --------

            - Adds levitation (#123)



            """
            ).lstrip(),
        )

    @with_project(
        config="""
        [tool.towncrier]
        name = "ImGoProject"
        """
    )
    def test_project_name_in_config(self, runner):
        """The calling towncrier with project name defined in configfile.

        Specifying a project name in toml file will be helpful to keep the
        project name consistent as part of the towncrier configuration, not call.
        """
        os.mkdir("newsfragments")
        with open("newsfragments/123.feature", "w") as f:
            f.write("Adds levitation")

        result = runner.invoke(
            _main, ["--version", "7.8.9", "--date", "01-01-2001", "--draft"]
        )

        self.assertEqual(0, result.exit_code, result.output)
        self.assertEqual(
            result.output,
            dedent(
                """
            Loading template...
            Finding news fragments...
            Rendering news fragments...
            Draft only -- nothing has been written.
            What is seen below is what would be written.

            ImGoProject 7.8.9 (01-01-2001)
            ==============================

            Features
            --------

            - Adds levitation (#123)



            """
            ).lstrip(),
        )

    @with_project(config="[tool.towncrier]")
    def test_no_package_changelog(self, runner):
        """The calling towncrier with any package argument.

        Specifying a package in the toml file or the command line
        should not always be needed:
        - we can set the version number on the command line,
          so we do not need the package for that.
        - we don't need to include the package in the changelog header.
        """
        os.mkdir("newsfragments")
        with open("newsfragments/123.feature", "w") as f:
            f.write("Adds levitation")

        result = runner.invoke(
            _main, ["--version", "7.8.9", "--date", "01-01-2001", "--draft"]
        )

        self.assertEqual(0, result.exit_code, result.output)
        self.assertEqual(
            result.output,
            dedent(
                """
            Loading template...
            Finding news fragments...
            Rendering news fragments...
            Draft only -- nothing has been written.
            What is seen below is what would be written.

            7.8.9 (01-01-2001)
            ==================

            Features
            --------

            - Adds levitation (#123)



            """
            ).lstrip(),
        )

    @with_project(
        config="""
        [tool.towncrier]
         single_file=false
         filename="{version}-notes.rst"
        """
    )
    def test_release_notes_in_separate_files(self, runner):
        """
        When `single_file = false` the release notes for each version are stored
        in a separate file.
        The name of the file is defined by the `filename` configuration value.
        """

        def do_build_once_with(version, fragment_file, fragment):
            with open(f"newsfragments/{fragment_file}", "w") as f:
                f.write(fragment)

            result = runner.invoke(
                _main,
                [
                    "--version",
                    version,
                    "--name",
                    "foo",
                    "--date",
                    "01-01-2001",
                    "--yes",
                ],
            )
            # Fragment files unknown to git are removed even without a git repo
            assert not Path(f"newsfragments/{fragment_file}").exists()
            return result

        results = []
        os.mkdir("newsfragments")
        results.append(do_build_once_with("7.8.9", "123.feature", "Adds levitation"))
        results.append(do_build_once_with("7.9.0", "456.bugfix", "Adds catapult"))

        self.assertEqual(0, results[0].exit_code, results[0].output)
        self.assertEqual(0, results[1].exit_code, results[1].output)
        self.assertEqual(
            2,
            len(list(Path.cwd().glob("*-notes.rst"))),
            "one newfile for each build",
        )
        self.assertTrue(os.path.exists("7.8.9-notes.rst"), os.listdir("."))
        self.assertTrue(os.path.exists("7.9.0-notes.rst"), os.listdir("."))

        outputs = []
        outputs.append(read("7.8.9-notes.rst"))
        outputs.append(read("7.9.0-notes.rst"))

        self.assertEqual(
            outputs[0],
            dedent(
                """
            foo 7.8.9 (01-01-2001)
            ======================

            Features
            --------

            - Adds levitation (#123)
            """
            ).lstrip(),
        )
        self.assertEqual(
            outputs[1],
            dedent(
                """
            foo 7.9.0 (01-01-2001)
            ======================

            Bugfixes
            --------

            - Adds catapult (#456)
            """
            ).lstrip(),
        )

    @with_project(
        config="""
        [tool.towncrier]
        singlefile="fail!"
        """
    )
    def test_singlefile_errors_and_explains_cleanly(self, runner):
        """
        Failure to find the configuration file results in a clean explanation
        without a traceback.
        """
        result = runner.invoke(_main)

        self.assertEqual(1, result.exit_code)
        self.assertEqual(
            "`singlefile` is not a valid option. Did you mean `single_file`?\n",
            result.output,
        )

    def test_all_version_notes_in_a_single_file(self):
        """
        When `single_file = true` the single file is used to store the notes
        for multiple versions.

        The name of the file is fixed as the literal option `filename` option
        in the configuration file, instead of extrapolated with variables.
        """
        runner = CliRunner()

        def do_build_once_with(version, fragment_file, fragment):
            with open(f"newsfragments/{fragment_file}", "w") as f:
                f.write(fragment)

            result = runner.invoke(
                _main,
                [
                    "--version",
                    version,
                    "--name",
                    "foo",
                    "--date",
                    "01-01-2001",
                    "--yes",
                ],
                catch_exceptions=False,
            )
            # Fragment files unknown to git are removed even without a git repo
            assert not Path(f"newsfragments/{fragment_file}").exists()
            return result

        results = []
        with runner.isolated_filesystem():
            with open("pyproject.toml", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "[tool.towncrier]",
                            " single_file=true",
                            " # The `filename` variable is fixed and not formated in any way.",
                            ' filename="{version}-notes.rst"',
                        ]
                    )
                )
            os.mkdir("newsfragments")
            results.append(
                do_build_once_with("7.8.9", "123.feature", "Adds levitation")
            )
            results.append(do_build_once_with("7.9.0", "456.bugfix", "Adds catapult"))

            self.assertEqual(0, results[0].exit_code, results[0].output)
            self.assertEqual(0, results[1].exit_code, results[1].output)
            self.assertEqual(
                1,
                len(list(Path.cwd().glob("*-notes.rst"))),
                "single newfile for multiple builds",
            )
            self.assertTrue(os.path.exists("{version}-notes.rst"), os.listdir("."))

            output = read("{version}-notes.rst")

            self.assertEqual(
                output,
                dedent(
                    """
                foo 7.9.0 (01-01-2001)
                ======================

                Bugfixes
                --------

                - Adds catapult (#456)


                foo 7.8.9 (01-01-2001)
                ======================

                Features
                --------

                - Adds levitation (#123)
                """
                ).lstrip(),
            )

    @with_project(
        config="""
        [tool.towncrier]
        template="towncrier:single-file-no-bullets"
        all_bullets=false
        """
    )
    def test_bullet_points_false(self, runner):
        """
        When all_bullets is false, subsequent lines are not indented.

        The automatic issue number inserted by towncrier will align with the
        manual bullet.
        """
        os.mkdir("newsfragments")
        with open("newsfragments/123.feature", "w") as f:
            f.write("wow!\n~~~~\n\nNo indentation at all.")
        with open("newsfragments/124.bugfix", "w") as f:
            f.write("#. Numbered bullet list.")
        with open("newsfragments/125.removal", "w") as f:
            f.write("- Hyphen based bullet list.")
        with open("newsfragments/126.doc", "w") as f:
            f.write("* Asterisk based bullet list.")

        result = runner.invoke(
            _main,
            [
                "--version",
                "7.8.9",
                "--name",
                "foo",
                "--date",
                "01-01-2001",
                "--yes",
            ],
        )

        self.assertEqual(0, result.exit_code, result.output)
        output = read("NEWS.rst")

        self.assertEqual(
            output,
            dedent(
                """
                foo 7.8.9 (01-01-2001)
                ======================

                Features
                --------

                wow!
                ~~~~

                No indentation at all.
                (#123)


                Bugfixes
                --------

                #. Numbered bullet list.
                   (#124)


                Improved Documentation
                ----------------------

                * Asterisk based bullet list.
                  (#126)


                Deprecations and Removals
                -------------------------

                - Hyphen based bullet list.
                  (#125)
                """
            ).lstrip(),
        )

    @with_project(
        config="""
        [tool.towncrier]
        package = "foo"
        title_format = "[{project_date}] CUSTOM RELEASE for {name} version {version}"
        """
    )
    def test_title_format_custom(self, runner):
        """
        A non-empty title format adds the specified title.
        """
        with open("foo/newsfragments/123.feature", "w") as f:
            f.write("Adds levitation")
        # Towncrier ignores .rst extension
        with open("foo/newsfragments/124.feature.rst", "w") as f:
            f.write("Extends levitation")

        result = runner.invoke(
            _main,
            [
                "--name",
                "FooBarBaz",
                "--version",
                "7.8.9",
                "--date",
                "20-01-2001",
                "--draft",
            ],
        )

        expected_output = dedent(
            """\
            Loading template...
            Finding news fragments...
            Rendering news fragments...
            Draft only -- nothing has been written.
            What is seen below is what would be written.

            [20-01-2001] CUSTOM RELEASE for FooBarBaz version 7.8.9
            =======================================================

            Features
            --------

            - Adds levitation (#123)
            - Extends levitation (#124)



        """
        )

        self.assertEqual(0, result.exit_code)
        self.assertEqual(expected_output, result.output)

    @with_project(
        config="""
        [tool.towncrier]
        package = "foo"
        filename = "NEWS.md"
        title_format = "[{project_date}] CUSTOM RELEASE for {name} version {version}"
        """
    )
    def test_title_format_custom_markdown(self, runner):
        """
        A non-empty title format adds the specified title, and if the target filename is
        markdown then the title is added as a markdown header.
        """
        with open("foo/newsfragments/123.feature", "w") as f:
            f.write("Adds levitation")
        result = runner.invoke(
            _main,
            [
                "--name",
                "FooBarBaz",
                "--version",
                "7.8.9",
                "--date",
                "20-01-2001",
                "--draft",
            ],
        )

        expected_output = dedent(
            """\
            Loading template...
            Finding news fragments...
            Rendering news fragments...
            Draft only -- nothing has been written.
            What is seen below is what would be written.

            # [20-01-2001] CUSTOM RELEASE for FooBarBaz version 7.8.9

            ### Features

            - Adds levitation (#123)



        """
        )

        self.assertEqual(0, result.exit_code)
        self.assertEqual(expected_output, result.output)

    @with_project(
        config="""
        [tool.towncrier]
        package = "foo"
        title_format = false
        template = "template.rst"
        """
    )
    def test_title_format_false(self, runner):
        """
        Setting the title format to false disables the explicit title.  This
        would be used, for example, when the template creates the title itself.
        """
        with open("template.rst", "w") as f:
            f.write(
                dedent(
                    """\
                Here's a hardcoded title added by the template
                ==============================================
                {% for section in sections %}
                {% set underline = "-" %}
                {% for category, val in definitions.items() if category in sections[section] %}

                {% for text, values in sections[section][category]|dictsort(by='value') %}
                - {{ text }}

                {% endfor %}
                {% endfor %}
                {% endfor %}
            """
                )
            )

        result = runner.invoke(
            _main,
            [
                "--name",
                "FooBarBaz",
                "--version",
                "7.8.9",
                "--date",
                "20-01-2001",
                "--draft",
            ],
            catch_exceptions=False,
        )

        expected_output = dedent(
            """\
            Loading template...
            Finding news fragments...
            Rendering news fragments...
            Draft only -- nothing has been written.
            What is seen below is what would be written.

            Here's a hardcoded title added by the template
            ==============================================

        """
        )

        self.assertEqual(0, result.exit_code)
        self.assertEqual(expected_output, result.output)

    @with_project(
        config="""
        [tool.towncrier]
        start_string="Release notes start marker"
        """
    )
    def test_start_string(self, runner):
        """
        The `start_string` configuration is used to detect the starting point
        for inserting the generated release notes. A newline is automatically
        added to the configured value.
        """
        os.mkdir("newsfragments")
        with open("newsfragments/123.feature", "w") as f:
            f.write("Adds levitation")
        with open("NEWS.rst", "w") as f:
            f.write("a line\n\nanother\n\nRelease notes start marker\na footer!\n")

        result = runner.invoke(
            _main,
            [
                "--version",
                "7.8.9",
                "--name",
                "foo",
                "--date",
                "01-01-2001",
                "--yes",
            ],
        )

        self.assertEqual(0, result.exit_code, result.output)
        self.assertTrue(os.path.exists("NEWS.rst"), os.listdir("."))
        output = read("NEWS.rst")

        expected_output = dedent(
            """\
            a line

            another

            Release notes start marker
            foo 7.8.9 (01-01-2001)
            ======================

            Features
            --------

            - Adds levitation (#123)


            a footer!
        """
        )

        self.assertEqual(expected_output, output)

    @with_project()
    def test_default_start_string(self, runner):
        """
        The default start string is ``.. towncrier release notes start``.
        """
        write("foo/newsfragments/123.feature", "Adds levitation")
        write(
            "NEWS.rst",
            contents="""
                a line

                another

                .. towncrier release notes start

                a footer!
            """,
            dedent=True,
        )

        result = runner.invoke(_main, ["--date", "01-01-2001"], catch_exceptions=False)
        self.assertEqual(0, result.exit_code, result.output)
        output = read("NEWS.rst")

        expected_output = dedent(
            """
            a line

            another

            .. towncrier release notes start

            Foo 1.2.3 (01-01-2001)
            ======================

            Features
            --------

            - Adds levitation (#123)


            a footer!
            """
        )

        self.assertEqual(expected_output, output)

    @with_project(
        config="""
        [tool.towncrier]
        package = "foo"
        filename = "NEWS.md"
        """
    )
    def test_default_start_string_markdown(self, runner):
        """
        The default start string is ``<!-- towncrier release notes start -->`` for
        Markdown.
        """
        write("foo/newsfragments/123.feature", "Adds levitation")
        write(
            "NEWS.md",
            contents="""
                a line

                another

                <!-- towncrier release notes start -->

                a footer!
            """,
            dedent=True,
        )

        result = runner.invoke(_main, ["--date", "01-01-2001"], catch_exceptions=False)
        self.assertEqual(0, result.exit_code, result.output)
        output = read("NEWS.md")

        expected_output = dedent(
            """
            a line

            another

            <!-- towncrier release notes start -->

            # Foo 1.2.3 (01-01-2001)

            ### Features

            - Adds levitation (#123)


            a footer!
            """
        )

        self.assertEqual(expected_output, output)

    @with_project(
        config="""
        [tool.towncrier]
        name = ""
        directory = "changes"
        filename = "NEWS.md"
        version = "1.2.3"
        """
    )
    def test_markdown_no_name_title(self, runner):
        """
        When configured with an empty `name` option,
        the default template used for Markdown
        renders the title of the release note with just
        the version number and release date.
        """
        write("changes/123.feature", "Adds levitation")
        write(
            "NEWS.md",
            contents="""
                A line

                <!-- towncrier release notes start -->
            """,
            dedent=True,
        )

        result = runner.invoke(_main, ["--date", "01-01-2001"], catch_exceptions=False)
        self.assertEqual(0, result.exit_code, result.output)
        output = read("NEWS.md")

        expected_output = dedent(
            """
            A line

            <!-- towncrier release notes start -->

            # 1.2.3 (01-01-2001)

            ### Features

            - Adds levitation (#123)
            """
        )

        self.assertEqual(expected_output, output)

    @with_project(
        config="""
        [tool.towncrier]
        title_format = "{version} - {project_date}"
        template = "template.rst"

          [[tool.towncrier.type]]
          directory = "feature"
          name = ""
          showcontent = true
        """
    )
    def test_with_topline_and_template_and_draft(self, runner):
        """
        Spacing is proper when drafting with a topline and a template.
        """
        os.mkdir("newsfragments")
        with open("newsfragments/123.feature", "w") as f:
            f.write("Adds levitation")
        with open("template.rst", "w") as f:
            f.write(
                dedent(
                    """\
                {% for section in sections %}
                {% set underline = "-" %}
                {% for category, val in definitions.items() if category in sections[section] %}

                {% for text, values in sections[section][category]|dictsort(by='value') %}
                - {{ text }}

                {% endfor %}
                {% endfor %}
                {% endfor %}
            """
                )
            )

        result = runner.invoke(
            _main,
            [
                "--version=7.8.9",
                "--name=foo",
                "--date=20-01-2001",
                "--draft",
            ],
        )

        expected_output = dedent(
            """\
            Loading template...
            Finding news fragments...
            Rendering news fragments...
            Draft only -- nothing has been written.
            What is seen below is what would be written.

            7.8.9 - 20-01-2001
            ==================

            - Adds levitation


        """
        )

        self.assertEqual(0, result.exit_code, result.output)
        self.assertEqual(expected_output, result.output)

    @with_project(
        config="""
        [tool.towncrier]
        """
    )
    def test_orphans_in_non_showcontent(self, runner):
        """
        When ``showcontent`` is false (like in the ``misc`` category by default),
        orphans are still rendered because they don't have an issue number to display.
        """
        os.mkdir("newsfragments")
        with open("newsfragments/123.misc", "w") as f:
            f.write("Misc")
        with open("newsfragments/345.misc", "w") as f:
            f.write("Another misc")
        with open("newsfragments/+.misc", "w") as f:
            f.write("Orphan misc still displayed!")
        with open("newsfragments/+2.misc", "w") as f:
            f.write("Another orphan misc still displayed!")

        result = runner.invoke(
            _main,
            [
                "--version=7.8.9",
                "--date=20-01-2001",
                "--draft",
            ],
        )

        expected_output = dedent(
            """\
            Loading template...
            Finding news fragments...
            Rendering news fragments...
            Draft only -- nothing has been written.
            What is seen below is what would be written.

            7.8.9 (20-01-2001)
            ==================

            Misc
            ----

            - #123, #345
            - Another orphan misc still displayed!
            - Orphan misc still displayed!



        """
        )

        self.assertEqual(0, result.exit_code, result.output)
        self.assertEqual(expected_output, result.output)

    @with_project(
        config="""
        [tool.towncrier]
        filename = "CHANGES.md"
        """
    )
    def test_orphans_in_non_showcontent_markdown(self, runner):
        """
        When ``showcontent`` is false (like in the ``misc`` category by default),
        orphans are still rendered because they don't have an issue number to display.
        """
        os.mkdir("newsfragments")
        with open("newsfragments/123.misc", "w") as f:
            f.write("Misc")
        with open("newsfragments/345.misc", "w") as f:
            f.write("Another misc")
        with open("newsfragments/+.misc", "w") as f:
            f.write("Orphan misc still displayed!")
        with open("newsfragments/+2.misc", "w") as f:
            f.write("Another orphan misc still displayed!")

        result = runner.invoke(
            _main,
            [
                "--version=7.8.9",
                "--date=20-01-2001",
                "--draft",
            ],
        )

        expected_output = dedent(
            """\
            Loading template...
            Finding news fragments...
            Rendering news fragments...
            Draft only -- nothing has been written.
            What is seen below is what would be written.

            # 7.8.9 (20-01-2001)

            ### Misc

            - #123, #345
            - Another orphan misc still displayed!
            - Orphan misc still displayed!



        """
        )

        self.assertEqual(0, result.exit_code, result.output)
        self.assertEqual(expected_output, result.output)

    @with_git_project()
    def test_uncommitted_files(self, runner, commit):
        """
        At build time, it will delete any fragment file regardless of its stage,
        included files that are not part of the git reporsitory,
        or are just staged or modified.
        """
        # 123 is committed, 124 is modified, 125 is just added, 126 is unknown

        with open("foo/newsfragments/123.feature", "w") as f:
            f.write("Adds levitation. File committed.")
        with open("foo/newsfragments/124.feature", "w") as f:
            f.write("Extends levitation. File modified in Git.")

        commit()

        with open("foo/newsfragments/125.feature", "w") as f:
            f.write("Baz levitation. Staged file.")
        with open("foo/newsfragments/126.feature", "w") as f:
            f.write("Fix (literal) crash. File unknown to Git.")

        with open("foo/newsfragments/124.feature", "a") as f:
            f.write(" Extended for an hour.")
        call(["git", "add", "foo/newsfragments/125.feature"])

        result = runner.invoke(_main, ["--date", "01-01-2001", "--yes"])

        self.assertEqual(0, result.exit_code)
        for fragment in ("123", "124", "125", "126"):
            self.assertFalse(os.path.isfile(f"foo/newsfragments/{fragment}.feature"))

        path = "NEWS.rst"
        self.assertTrue(os.path.isfile(path))
        news_contents = open(path).read()
        self.assertEqual(
            news_contents,
            dedent(
                """\
                Foo 1.2.3 (01-01-2001)
                ======================

                Features
                --------

                - Adds levitation. File committed. (#123)
                - Extends levitation. File modified in Git. Extended for an hour. (#124)
                - Baz levitation. Staged file. (#125)
                - Fix (literal) crash. File unknown to Git. (#126)
                """
            ),
        )

    @with_project(
        config="""
        [tool.towncrier]
        package = "foo"
        ignore = ["template.jinja", "CAPYBARAS.md"]
        """
    )
    def test_ignored_files(self, runner):
        """
        When `ignore` is set in config, files with those names are ignored.
        """
        with open("foo/newsfragments/123.feature", "w") as f:
            f.write("This has valid filename (control case)")
        with open("foo/newsfragments/template.jinja", "w") as f:
            f.write("This template has been manually ignored")
        with open("foo/newsfragments/CAPYBARAS.md", "w") as f:
            f.write("This markdown file has been manually ignored")
        with open("foo/newsfragments/.gitignore", "w") as f:
            f.write("gitignore is automatically ignored")

        result = runner.invoke(
            _main, ["--draft", "--date", "01-01-2001", "--version", "1.0.0"]
        )
        self.assertEqual(0, result.exit_code, result.output)

    @with_project(
        config="""
        [tool.towncrier]
        package = "foo"
        ignore = []
        """
    )
    def test_invalid_fragment_name(self, runner):
        """
        When `ignore` is set in config, invalid filenames cause failure.
        """
        with open("foo/newsfragments/123.feature", "w") as f:
            f.write("This has valid filename (control case)")
        with open("foo/newsfragments/feature.124", "w") as f:
            f.write("This has the issue and category the wrong way round")

        result = runner.invoke(
            _main, ["--draft", "--date", "01-01-2001", "--version", "1.0.0"]
        )
        self.assertEqual(1, result.exit_code, result.output)
        self.assertIn("Invalid news fragment name: feature.124", result.output)

    @with_project()
    def test_no_ignore_configured(self, runner):
        """
        When `ignore` is not set in config, invalid filenames are skipped.

        This maintains backward compatibility with before we added `ignore`
        to the configuration spec.
        """
        with open("foo/newsfragments/feature.124", "w") as f:
            f.write("This has the issue and category the wrong way round")

        result = runner.invoke(
            _main, ["--draft", "--date", "01-01-2001", "--version", "1.0.0"]
        )
        self.assertEqual(0, result.exit_code, result.output)



# Copyright (c) Amber Brown, 2017
# See LICENSE for details.

import os
import os.path
import warnings

from pathlib import Path
from subprocess import call

from click.testing import CliRunner
from twisted.trial.unittest import TestCase

from towncrier import check
from towncrier.build import _main as towncrier_build
from towncrier.check import _main as towncrier_check

from .helpers import setup_simple_project, with_isolated_runner, write


def create_project(
    pyproject_path="pyproject.toml", main_branch="main", extra_config=""
):
    """
    Create the project files in the main branch that already has a
    news-fragment and then switch to a new in-work branch.
    """
    setup_simple_project(pyproject_path=pyproject_path, extra_config=extra_config)
    Path("foo/newsfragments/123.feature").write_text("Adds levitation")
    initial_commit(branch=main_branch)
    call(["git", "checkout", "-b", "otherbranch"])


def commit(message):
    """Stage and commit the repo in the current working directory

    There must be uncommitted changes otherwise git will complain:
    "nothing to commit, working tree clean"
    """
    call(["git", "add", "."])
    call(["git", "commit", "-m", message])


def initial_commit(branch="main"):
    """
    Create a git repo, configure it and make an initial commit

    There must be uncommitted changes otherwise git will complain:
    "nothing to commit, working tree clean"
    """
    # --initial-branch is explicitly set to `main` because
    # git has deprecated the default branch name.
    call(["git", "init", f"--initial-branch={branch}"])
    # Without ``git config` user.name and user.email `git commit` fails
    # unless the settings are set globally
    call(["git", "config", "user.name", "user"])
    call(["git", "config", "user.email", "user@example.com"])
    commit("Initial Commit")


class TestChecker(TestCase):
    maxDiff = None

    def test_git_fails(self):
        """
        If git fails to report a comparison, git's output is reported to aid in
        debugging the situation.
        """
        runner = CliRunner()
        with runner.isolated_filesystem():
            create_project("pyproject.toml")

            result = runner.invoke(towncrier_check, ["--compare-with", "hblaugh"])
            self.assertIn("git produced output while failing", result.output)
            self.assertIn("hblaugh", result.output)

    def test_no_changes_made(self):
        self._test_no_changes_made(
            "pyproject.toml", lambda runner, main, argv: runner.invoke(main, argv)
        )

    def test_no_changes_made_config_path(self):
        pyproject = "not-pyproject.toml"
        self._test_no_changes_made(
            pyproject,
            lambda runner, main, argv: runner.invoke(
                main, argv + ["--config", pyproject]
            ),
        )

    def _test_no_changes_made(self, pyproject_path, invoke):
        """
        When no changes are made on a new branch, no checks are performed.
        """
        runner = CliRunner()

        with runner.isolated_filesystem():
            create_project(pyproject_path, main_branch="master")

            result = invoke(runner, towncrier_check, ["--compare-with", "master"])

            self.assertEqual(0, result.exit_code, result.output)
            self.assertEqual(
                "On master branch, or no diffs, so no newsfragment required.\n",
                result.output,
            )

    @with_isolated_runner
    def test_fragment_exists(self, runner):
        create_project("pyproject.toml")

        write("foo/somefile.py", "import os")
        commit("add a file")

        fragment_path = Path("foo/newsfragments/1234.feature").absolute()
        write(fragment_path, "Adds gravity back")
        commit("add a newsfragment")

        result = runner.invoke(towncrier_check, ["--compare-with", "main"])

        self.assertTrue(
            result.output.endswith("Found:\n1. " + str(fragment_path) + "\n"),
            (result.output, str(fragment_path)),
        )
        self.assertEqual(0, result.exit_code, result.output)

    @with_isolated_runner
    def test_fragment_exists_hidden(self, runner):
        """
        Location of fragments can be configured using tool.towncrier.directory.
        """
        create_project("pyproject.toml", extra_config="directory = 'deep/fragz'\n")

        write("foo/bar/somefile.py", "import os")
        commit("add a file")

        fragment_path = Path("deep/fragz/1234.feature").absolute()
        write(fragment_path, "Adds gravity back")
        commit("add a newsfragment")

        result = runner.invoke(towncrier_check, ["--compare-with", "main"])

        self.assertTrue(
            result.output.endswith("Found:\n1. " + str(fragment_path) + "\n"),
            (result.output, str(fragment_path)),
        )
        self.assertEqual(0, result.exit_code, result.output)

    def test_fragment_missing(self):
        runner = CliRunner()

        with runner.isolated_filesystem():
            create_project("pyproject.toml", main_branch="master")

            file_path = "foo/somefile.py"
            with open(file_path, "w") as f:
                f.write("import os")

            call(["git", "add", "foo/somefile.py"])
            call(["git", "commit", "-m", "add a file"])

            result = runner.invoke(towncrier_check, ["--compare-with", "master"])

            self.assertEqual(1, result.exit_code)
            self.assertTrue(
                result.output.endswith("No new newsfragments found on this branch.\n")
            )

    def test_fragment_exists_but_not_in_check(self):
        """A fragment that exists but is marked as check=False is ignored by the check."""
        runner = CliRunner()

        with runner.isolated_filesystem():
            create_project(
                "pyproject.toml",
                main_branch="master",
                extra_config="[[tool.towncrier.type]]\n"
                'directory = "feature"\n'
                'name = "Features"\n'
                "showcontent = true\n"
                "[[tool.towncrier.type]]\n"
                'directory = "sut"\n'
                'name = "System Under Test"\n'
                "showcontent = true\n"
                "check=false\n",
            )

            file_path = "foo/somefile.py"
            write(file_path, "import os")

            fragment_path = Path("foo/newsfragments/1234.sut").absolute()
            write(fragment_path, "Adds gravity back")
            commit("add a file and a newsfragment")

            result = runner.invoke(towncrier_check, ["--compare-with", "master"])

            self.assertEqual(1, result.exit_code)
            self.assertTrue(
                result.output.endswith(
                    "Found newsfragments of unchecked types in the branch:\n1. "
                    + str(fragment_path)
                    + "\n"
                ),
                (result.output, str(fragment_path)),
            )

    def test_fragment_exists_and_in_check(self):
        """
        A fragment that exists but is not marked as check=False is
        not ignored by the check, even if other categories are marked as check=False.
        """
        runner = CliRunner()

        with runner.isolated_filesystem():
            create_project(
                "pyproject.toml",
                main_branch="master",
                extra_config="[[tool.towncrier.type]]\n"
                'directory = "feature"\n'
                'name = "Features"\n'
                "showcontent = true\n"
                "[[tool.towncrier.type]]\n"
                'directory = "sut"\n'
                'name = "System Under Test"\n'
                "showcontent = true\n"
                "check=false\n",
            )

            file_path = "foo/somefile.py"
            write(file_path, "import os")

            fragment_path = Path("foo/newsfragments/1234.feature").absolute()
            write(fragment_path, "Adds gravity back")
            commit("add a file and a newsfragment")

            result = runner.invoke(towncrier_check, ["--compare-with", "master"])

            self.assertEqual(0, result.exit_code)
            self.assertTrue(
                result.output.endswith("Found:\n1. " + str(fragment_path) + "\n"),
                (result.output, str(fragment_path)),
            )

    def test_none_stdout_encoding_works(self):
        """
        No failure when output is piped causing None encoding for stdout.
        """
        runner = CliRunner()

        with runner.isolated_filesystem():
            create_project("pyproject.toml", main_branch="master")

            fragment_path = "foo/newsfragments/1234.feature"
            with open(fragment_path, "w") as f:
                f.write("Adds gravity back")

            call(["git", "add", fragment_path])
            call(["git", "commit", "-m", "add a newsfragment"])

            runner = CliRunner(mix_stderr=False)
            result = runner.invoke(towncrier_check, ["--compare-with", "master"])

        self.assertEqual(0, result.exit_code)
        self.assertEqual(0, len(result.stderr))

    def test_first_release(self):
        """
        The checks should be skipped on a branch that creates the news file.

        If the checks are not skipped in this case, towncrier check would fail
        for the first release that has a changelog.
        """
        runner = CliRunner()

        with runner.isolated_filesystem():
            # Arrange
            create_project()
            # Before any release, the NEWS file might no exist.
            self.assertNotIn("NEWS.rst", os.listdir("."))

            runner.invoke(towncrier_build, ["--yes", "--version", "1.0"])
            commit("Prepare a release")
            # When missing,
            # the news file is automatically created with a new release.
            self.assertIn("NEWS.rst", os.listdir("."))

            # Act
            result = runner.invoke(towncrier_check, ["--compare-with", "main"])

            # Assert
            self.assertEqual(0, result.exit_code, (result, result.output))
            self.assertIn("Checks SKIPPED: news file changes detected", result.output)

    def test_release_branch(self):
        """
        The checks for missing news fragments are skipped on a branch that
        modifies the news file.
        This is a hint that we are on a release branch
        and at release time is expected no not have news-fragment files.
        """
        runner = CliRunner()

        with runner.isolated_filesystem():
            # Arrange
            create_project()

            # Do a first release without any checks.
            # And merge the release branch back into the main branch.
            runner.invoke(towncrier_build, ["--yes", "--version", "1.0"])
            commit("First release")
            # The news file is now created.
            self.assertIn("NEWS.rst", os.listdir("."))
            call(["git", "checkout", "main"])
            call(["git", "merge", "otherbranch", "-m", "Sync release in main branch."])

            # We have a new feature branch that has a news fragment that
            # will be merged to the main branch.
            call(["git", "checkout", "-b", "new-feature-branch"])
            write("foo/newsfragments/456.feature", "Foo the bar")
            commit("A feature in the second release.")
            call(["git", "checkout", "main"])
            call(
                [
                    "git",
                    "merge",
                    "new-feature-branch",
                    "-m",
                    "Merge new-feature-branch.",
                ]
            )

            # We now have the new release branch.
            call(["git", "checkout", "-b", "next-release"])
            runner.invoke(towncrier_build, ["--yes", "--version", "2.0"])
            commit("Second release")

            # Act
            result = runner.invoke(towncrier_check, ["--compare-with", "main"])

            # Assert
            self.assertEqual(0, result.exit_code, (result, result.output))
            self.assertIn("Checks SKIPPED: news file changes detected", result.output)

    def test_get_default_compare_branch_missing(self):
        """
        If there's no recognized remote origin, exit with an error.
        """
        runner = CliRunner()

        with runner.isolated_filesystem():
            create_project()

            result = runner.invoke(towncrier_check)

        self.assertEqual(1, result.exit_code)
        self.assertEqual("Could not detect default branch. Aborting.\n", result.output)

    def test_get_default_compare_branch_main(self):
        """
        If there's a remote branch origin/main, prefer it over everything else.
        """
        branch = check._get_default_compare_branch(["origin/master", "origin/main"])

        self.assertEqual("origin/main", branch)

    def test_get_default_compare_branch_fallback(self):
        """
        If there's origin/master and no main, use it and warn about it.
        """
        with warnings.catch_warnings(record=True) as w:
            branch = check._get_default_compare_branch(["origin/master", "origin/foo"])

        self.assertEqual("origin/master", branch)
        self.assertTrue(w[0].message.args[0].startswith('Using "origin/master'))

    @with_isolated_runner
    def test_in_different_dir_with_nondefault_newsfragments_directory(self, runner):
        """
        It can check the fragments located in a sub-directory
        that is specified using the `--dir` CLI argument.
        """
        main_branch = "main"
        Path("pyproject.toml").write_text(
            # Important to customize `config.directory` because the default
            # already supports this scenario.
            "[tool.towncrier]\n"
            + 'directory = "changelog.d"\n'
        )
        subproject1 = Path("foo")
        (subproject1 / "foo").mkdir(parents=True)
        (subproject1 / "foo/__init__.py").write_text("")
        (subproject1 / "changelog.d").mkdir(parents=True)
        (subproject1 / "changelog.d/123.feature").write_text("Adds levitation")
        initial_commit(branch=main_branch)
        call(["git", "checkout", "-b", "otherbranch"])

        # We add a code change but forget to add a news fragment.
        write(subproject1 / "foo/somefile.py", "import os")
        commit("add a file")
        result = runner.invoke(
            towncrier_check,
            (
                "--config",
                "pyproject.toml",
                "--dir",
                str(subproject1),
                "--compare-with",
                "main",
            ),
        )

        self.assertEqual(1, result.exit_code)
        self.assertTrue(
            result.output.endswith("No new newsfragments found on this branch.\n")
        )

        # We add the news fragment.
        fragment_path = (subproject1 / "changelog.d/124.feature").absolute()
        write(fragment_path, "Adds gravity back")
        commit("add a newsfragment")
        result = runner.invoke(
            towncrier_check,
            ("--config", "pyproject.toml", "--dir", "foo", "--compare-with", "main"),
        )

        self.assertEqual(0, result.exit_code, result.output)
        self.assertTrue(
            result.output.endswith("Found:\n1. " + str(fragment_path) + "\n"),
            (result.output, str(fragment_path)),
        )

        # We add a change in a different subproject without a news fragment.
        # Checking subproject1 should pass.
        subproject2 = Path("bar")
        (subproject2 / "bar").mkdir(parents=True)
        (subproject2 / "changelog.d").mkdir(parents=True)
        write(subproject2 / "bar/somefile.py", "import os")
        commit("add a file")
        result = runner.invoke(
            towncrier_check,
            (
                "--config",
                "pyproject.toml",
                "--dir",
                subproject1,
                "--compare-with",
                "main",
            ),
        )

        self.assertEqual(0, result.exit_code, result.output)
        self.assertTrue(
            result.output.endswith("Found:\n1. " + str(fragment_path) + "\n"),
            (result.output, str(fragment_path)),
        )

        # Checking subproject2 should result in an error.
        result = runner.invoke(
            towncrier_check,
            (
                "--config",
                "pyproject.toml",
                "--dir",
                subproject2,
                "--compare-with",
                "main",
            ),
        )
        self.assertEqual(1, result.exit_code)
        self.assertTrue(
            result.output.endswith("No new newsfragments found on this branch.\n")
        )

    @with_isolated_runner
    def test_ignored_files(self, runner):
        """
        When `ignore` is set in config, files with those names are ignored.
        """
        create_project("pyproject.toml", extra_config='ignore = ["template.jinja"]')

        write(
            "foo/newsfragments/124.feature",
            "This fragment has valid name (control case)",
        )
        write("foo/newsfragments/template.jinja", "This is manually ignored")
        write("foo/newsfragments/.gitignore", "gitignore is automatically ignored")
        commit("add stuff")

        result = runner.invoke(towncrier_check, ["--compare-with", "main"])
        self.assertEqual(0, result.exit_code, result.output)

    @with_isolated_runner
    def test_invalid_fragment_name(self, runner):
        """
        Fails if a news fragment has an invalid name, even if `ignore` is not set in
        the config.
        """
        create_project("pyproject.toml")

        write(
            "foo/newsfragments/124.feature",
            "This fragment has valid name (control case)",
        )
        write(
            "foo/newsfragments/feature.125",
            "This has issue and category wrong way round",
        )
        write(
            "NEWS.rst",
            "Modification of news file should not skip check of invalid names",
        )
        commit("add stuff")

        result = runner.invoke(towncrier_check, ["--compare-with", "main"])
        self.assertEqual(1, result.exit_code, result.output)
        self.assertIn("Invalid news fragment name: feature.125", result.output)



# Copyright (c) Stephen Finucane, 2019
# See LICENSE for details.

"""
Entry point of the command line interface.

Each sub-command has its separate CLI definition andd help messages.
"""

from __future__ import annotations

import click

from ._version import __version__
from .build import _main as _build_cmd
from .check import _main as _check_cmd
from .click_default_group import DefaultGroup
from .create import _main as _create_cmd


@click.group(cls=DefaultGroup, default="build", default_if_no_args=True)
@click.version_option(__version__)
def cli() -> None:
    """
    Towncrier is a utility to produce useful, summarised news files for your project.
    Rather than reading the Git history as some newer tools to produce it, or having
    one single file which developers all write to, towncrier reads "news fragments"
    which contain information useful to end users.

    Towncrier delivers the news which is convenient to those that hear it, not those
    that write it.

    That is, a “news fragment” (a small file containing just enough information to
    be useful to end users) can be written that summarises what has changed from the
    “developer log” (which may contain complex information about the original issue,
    how it was fixed, who authored the fix, and who reviewed the fix). By compiling
    a collection of these fragments, towncrier can produce a digest of the changes
    which is valuable to those who may wish to use the software.
    """
    pass


cli.add_command(_build_cmd)
cli.add_command(_check_cmd)
cli.add_command(_create_cmd)



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.


from twisted.trial.unittest import TestCase

from .._builder import render_fragments, split_fragments
from .helpers import read_pkg_resource


class FormatterTests(TestCase):
    maxDiff = None

    def test_split(self):
        fragments = {
            "": {
                ("1", "misc", 0): "",
                ("baz", "misc", 0): "",
                ("2", "feature", 0): "Foo added.",
                ("5", "feature", 0): "Foo added.    \n",
                ("6", "bugfix", 0): "Foo added.",
            },
            "Web": {
                ("3", "bugfix", 0): "Web fixed.    ",
                ("4", "feature", 0): "Foo added.",
            },
        }

        expected_output = {
            "": {
                "misc": {"": ["1", "baz"]},
                "feature": {"Foo added.": ["2", "5"]},
                "bugfix": {"Foo added.": ["6"]},
            },
            "Web": {
                "bugfix": {"Web fixed.": ["3"]},
                "feature": {"Foo added.": ["4"]},
            },
        }

        definitions = {
            "feature": {"name": "Features", "showcontent": True},
            "bugfix": {"name": "Bugfixes", "showcontent": True},
            "misc": {"name": "Misc", "showcontent": False},
        }

        output = split_fragments(fragments, definitions)

        self.assertEqual(expected_output, output)

    def test_basic(self):
        """
        Basic functionality -- getting a bunch of news fragments and formatting
        them into a rST file -- works.
        """
        fragments = {
            "": {
                # asciibetical sorting will do 1, 142, 9
                # we want 1, 9, 142 instead
                ("142", "misc", 0): "",
                ("1", "misc", 0): "",
                ("9", "misc", 0): "",
                ("bar", "misc", 0): "",
                ("4", "feature", 0): "Stuff!",
                ("2", "feature", 0): "Foo added.",
                ("72", "feature", 0): "Foo added.",
                ("9", "feature", 0): "Foo added.",
                ("baz", "feature", 0): "Fun!",
            },
            "Names": {},
            "Web": {("3", "bugfix", 0): "Web fixed."},
        }

        definitions = {
            "feature": {"name": "Features", "showcontent": True},
            "bugfix": {"name": "Bugfixes", "showcontent": True},
            "misc": {"name": "Misc", "showcontent": False},
        }

        expected_output = """MyProject 1.0 (never)
=====================

Features
--------

- Fun! (baz)
- Foo added. (#2, #9, #72)
- Stuff! (#4)


Misc
----

- bar, #1, #9, #142


Names
-----

No significant changes.


Web
---

Bugfixes
~~~~~~~~

- Web fixed. (#3)


"""

        template = read_pkg_resource("templates/default.rst")

        fragments = split_fragments(fragments, definitions)
        output = render_fragments(
            template,
            None,
            fragments,
            definitions,
            ["-", "~"],
            wrap=True,
            versiondata={"name": "MyProject", "version": "1.0", "date": "never"},
        )
        self.assertEqual(output, expected_output)

        # Check again with non-default underlines
        expected_output_weird_underlines = """MyProject 1.0 (never)
=====================

Features
********

- Fun! (baz)
- Foo added. (#2, #9, #72)
- Stuff! (#4)


Misc
****

- bar, #1, #9, #142


Names
*****

No significant changes.


Web
***

Bugfixes
^^^^^^^^

- Web fixed. (#3)


"""

        output = render_fragments(
            template,
            None,
            fragments,
            definitions,
            ["*", "^"],
            wrap=True,
            versiondata={"name": "MyProject", "version": "1.0", "date": "never"},
        )
        self.assertEqual(output, expected_output_weird_underlines)

    def test_markdown(self):
        """
        Check formating of default markdown template.
        """
        fragments = {
            "": {
                # asciibetical sorting will do 1, 142, 9
                # we want 1, 9, 142 instead
                ("142", "misc", 0): "",
                ("1", "misc", 0): "",
                ("9", "misc", 0): "",
                ("bar", "misc", 0): "",
                ("4", "feature", 0): "Stuff!",
                ("2", "feature", 0): "Foo added.",
                ("72", "feature", 0): "Foo added.",
                ("9", "feature", 0): "Foo added.",
                ("3", "feature", 0): "Multi-line\nhere",
                ("baz", "feature", 0): "Fun!",
            },
            "Names": {},
            "Web": {
                ("3", "bugfix", 0): "Web fixed.",
                ("2", "bugfix", 0): "Multi-line bulleted\n- fix\n- here",
            },
        }

        definitions = {
            "feature": {"name": "Features", "showcontent": True},
            "bugfix": {"name": "Bugfixes", "showcontent": True},
            "misc": {"name": "Misc", "showcontent": False},
        }

        expected_output = """# MyProject 1.0 (never)

### Features

- Fun! (baz)
- Foo added. (#2, #9, #72)
- Multi-line
  here (#3)
- Stuff! (#4)

### Misc

- bar, #1, #9, #142


## Names

No significant changes.


## Web

### Bugfixes

- Multi-line bulleted
  - fix
  - here

  (#2)
- Web fixed. (#3)


"""

        template = read_pkg_resource("templates/default.md")

        fragments = split_fragments(fragments, definitions)
        output = render_fragments(
            template,
            None,
            fragments,
            definitions,
            ["-", "~"],
            wrap=True,
            versiondata={"name": "MyProject", "version": "1.0", "date": "never"},
        )
        self.assertEqual(output, expected_output)

        # Also test with custom issue format
        expected_output = """# MyProject 1.0 (never)

### Features

- Fun! ([baz])
- Foo added. ([2], [9], [72])
- Multi-line
  here ([3])
- Stuff! ([4])

[baz]: https://github.com/twisted/towncrier/issues/baz
[2]: https://github.com/twisted/towncrier/issues/2
[3]: https://github.com/twisted/towncrier/issues/3
[4]: https://github.com/twisted/towncrier/issues/4
[9]: https://github.com/twisted/towncrier/issues/9
[72]: https://github.com/twisted/towncrier/issues/72

### Misc

- [bar], [1], [9], [142]

[bar]: https://github.com/twisted/towncrier/issues/bar
[1]: https://github.com/twisted/towncrier/issues/1
[9]: https://github.com/twisted/towncrier/issues/9
[142]: https://github.com/twisted/towncrier/issues/142


## Names

No significant changes.


## Web

### Bugfixes

- Multi-line bulleted
  - fix
  - here

  ([2])
- Web fixed. ([3])

[2]: https://github.com/twisted/towncrier/issues/2
[3]: https://github.com/twisted/towncrier/issues/3


"""

        output = render_fragments(
            template,
            "[{issue}]: https://github.com/twisted/towncrier/issues/{issue}",
            fragments,
            definitions,
            ["-", "~"],
            wrap=True,
            versiondata={"name": "MyProject", "version": "1.0", "date": "never"},
        )

        self.assertEqual(output, expected_output)

    def test_issue_format(self):
        """
        issue_format option can be used to format issue text.
        And sorting happens before formatting, so numerical issues are still
        ordered numerically even if that doesn't match asciibetical order on
        the final text.
        """
        fragments = {
            "": {
                # asciibetical sorting will do 1, 142, 9
                # we want 1, 9, 142 instead
                ("142", "misc", 0): "",
                ("1", "misc", 0): "",
                ("9", "misc", 0): "",
                ("bar", "misc", 0): "",
            }
        }

        definitions = {"misc": {"name": "Misc", "showcontent": False}}

        expected_output = """MyProject 1.0 (never)
=====================

Misc
----

- xxbar, xx1, xx9, xx142


"""

        template = read_pkg_resource("templates/default.rst")

        fragments = split_fragments(fragments, definitions)
        output = render_fragments(
            template,
            "xx{issue}",
            fragments,
            definitions,
            ["-", "~"],
            wrap=True,
            versiondata={"name": "MyProject", "version": "1.0", "date": "never"},
        )
        self.assertEqual(output, expected_output)

    def test_line_wrapping(self):
        """
        Output is nicely wrapped, but doesn't break up words (which can mess
        up URLs)
        """
        self.maxDiff = None

        fragments = {
            "": {
                (
                    "1",
                    "feature",
                    0,
                ): """
                asdf asdf asdf asdf looooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooong newsfragment.
                """,  # NOQA
                ("2", "feature", 0): "https://google.com/q=?" + "-" * 100,
                ("3", "feature", 0): "a " * 80,
            }
        }

        definitions = {"feature": {"name": "Features", "showcontent": True}}

        expected_output = """MyProject 1.0 (never)
=====================

Features
--------

- asdf asdf asdf asdf
  looooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooong
  newsfragment. (#1)
-
  https://google.com/q=?----------------------------------------------------------------------------------------------------
  (#2)
- a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a
  a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a
  a a (#3)


"""

        template = read_pkg_resource("templates/default.rst")

        fragments = split_fragments(fragments, definitions)
        output = render_fragments(
            template,
            None,
            fragments,
            definitions,
            ["-", "~"],
            wrap=True,
            versiondata={"name": "MyProject", "version": "1.0", "date": "never"},
        )
        self.assertEqual(output, expected_output)

    def test_line_wrapping_disabled(self):
        """
        Output is not wrapped if it's disabled.
        """
        self.maxDiff = None

        fragments = {
            "": {
                (
                    "1",
                    "feature",
                    0,
                ): """
                asdf asdf asdf asdf looooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooong newsfragment.
                """,  # NOQA
                ("2", "feature", 0): "https://google.com/q=?" + "-" * 100,
                ("3", "feature", 0): "a " * 80,
            }
        }

        definitions = {"feature": {"name": "Features", "showcontent": True}}

        expected_output = """MyProject 1.0 (never)
=====================

Features
--------

- asdf asdf asdf asdf looooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooong newsfragment. (#1)
- https://google.com/q=?---------------------------------------------------------------------------------------------------- (#2)
- a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a a (#3)


"""  # NOQA

        template = read_pkg_resource("templates/default.rst")

        fragments = split_fragments(fragments, definitions)
        output = render_fragments(
            template,
            None,
            fragments,
            definitions,
            ["-", "~"],
            wrap=False,
            versiondata={"name": "MyProject", "version": "1.0", "date": "never"},
        )
        self.assertEqual(output, expected_output)



from __future__ import annotations

import abc

from typing import Any, Iterable, Mapping


class BaseFragmentTypesLoader:
    """Base class to load fragment types."""

    __metaclass__ = abc.ABCMeta

    def __init__(self, config: Mapping[str, Any]):
        """Initialize."""
        self.config = config

    @classmethod
    def factory(cls, config: Mapping[str, Any]) -> BaseFragmentTypesLoader:
        fragment_types_class: type[BaseFragmentTypesLoader] = DefaultFragmentTypesLoader
        fragment_types = config.get("fragment", {})
        types_config = config.get("type", {})
        if fragment_types:
            fragment_types_class = TableFragmentTypesLoader
        elif types_config:
            fragment_types_class = ArrayFragmentTypesLoader

        new = fragment_types_class(config)
        return new

    @abc.abstractmethod
    def load(self) -> Mapping[str, Mapping[str, Any]]:
        """Load fragment types."""


class DefaultFragmentTypesLoader(BaseFragmentTypesLoader):
    """Default towncrier's fragment types."""

    _default_types = {
        # Keep in-sync with docs/tutorial.rst.
        "feature": {"name": "Features", "showcontent": True, "check": True},
        "bugfix": {"name": "Bugfixes", "showcontent": True, "check": True},
        "doc": {"name": "Improved Documentation", "showcontent": True, "check": True},
        "removal": {
            "name": "Deprecations and Removals",
            "showcontent": True,
            "check": True,
        },
        "misc": {"name": "Misc", "showcontent": False, "check": True},
    }

    def load(self) -> Mapping[str, Mapping[str, Any]]:
        """Load default types."""
        return self._default_types


class ArrayFragmentTypesLoader(BaseFragmentTypesLoader):
    """Load fragment types from an toml array of tables.

    This loader get the custom fragment types defined through a
    toml array of tables, that ``toml`` parses as an array
    of mappings.

    For example::

        [tool.towncrier]
        [[tool.towncrier.type]]
        directory = "deprecation"
        name = "Deprecations"
        showcontent = true

    """

    def load(self) -> Mapping[str, Mapping[str, Any]]:
        """Load types from toml array of mappings."""

        types = {}
        types_config = self.config["type"]
        for type_config in types_config:
            directory = type_config["directory"]
            fragment_type_name = type_config["name"]
            is_content_required = type_config["showcontent"]
            check = type_config.get("check", True)
            types[directory] = {
                "name": fragment_type_name,
                "showcontent": is_content_required,
                "check": check,
            }
        return types


class TableFragmentTypesLoader(BaseFragmentTypesLoader):
    """Load fragment types from toml tables.

    This loader get the custom fragment types defined through a
    toml  tables, that ``toml`` parses as an nested mapping.

    This loader allows omitting ``name`` and
    ```showcontent`` fields.
    ``name`` by default is the capitalized
    fragment type.
    ``showcontent`` is true by default.

    For example::

        [tool.towncrier]
        [tool.towncrier.fragment.chore]
        name = "Chores"
        showcontent = False

        [tool.towncrier.fragment.deprecations]
        # name will be "Deprecations"
        # The content will be shown.

    """

    def __init__(self, config: Mapping[str, Mapping[str, Any]]):
        """Initialize."""
        self.config = config
        self.fragment_options = config.get("fragment", {})

    def load(self) -> Mapping[str, Mapping[str, Any]]:
        """Load types from nested mapping."""
        fragment_types: Iterable[str] = self.fragment_options.keys()
        fragment_types = sorted(fragment_types)
        custom_types_sequence = [
            (fragment_type, self._load_options(fragment_type))
            for fragment_type in fragment_types
        ]
        types = dict(custom_types_sequence)
        return types

    def _load_options(self, fragment_type: str) -> Mapping[str, Any]:
        """Load fragment options."""
        capitalized_fragment_type = fragment_type.capitalize()
        options = self.fragment_options.get(fragment_type, {})
        fragment_description = options.get("name", capitalized_fragment_type)
        show_content = options.get("showcontent", True)
        check = options.get("check", True)
        clean_fragment_options = {
            "name": fragment_description,
            "showcontent": show_content,
            "check": check,
        }
        return clean_fragment_options



# Copyright (c) Amber Brown, 2015
# See LICENSE for details.

from twisted.trial.unittest import TestCase

import towncrier


class TestPackaging(TestCase):
    def test_version_attr(self):
        """
        towncrier.__version__ was deprecated, but still exists for now.
        """

        def access__version():
            return towncrier.__version__

        expected_warning = (
            "Accessing towncrier.__version__ is deprecated and will be "
            "removed in a future release. Use importlib.metadata directly "
            "to query for towncrier's packaging metadata."
        )

        self.assertWarns(
            DeprecationWarning, expected_warning, __file__, access__version
        )
